# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from os import sep
from pathlib import Path
from subprocess import Popen
import sys
from sys import stdout
from time import (sleep,
                  time)
from traceback import print_exc

from Nagstamon.config import (AppInfo,
                              conf,
                              DESKTOP_WAYLAND,
                              OS,
                              OS_MACOS,
                              OS_NON_LINUX,
                              OS_WINDOWS,
                              RESOURCES, debug_queue)
from Nagstamon.helpers import (STATES,
                               STATES_SOUND)
from Nagstamon.qui.constants import WINDOW_FLAGS
from Nagstamon.qui.globals import (dbus_connection,
                                   statuswindow_properties)
from Nagstamon.qui.helpers import (get_screen_geometry,
                                   get_screen_name,
                                   hide_macos_dock_icon)
from Nagstamon.qui.qt import (QAction,
                              QCursor,
                              QIcon,
                              QMenuBar,
                              QMessageBox,
                              QObject,
                              QVBoxLayout,
                              QScrollArea,
                              QSizePolicy,
                              QSpacerItem,
                              QThread,
                              QWidget,
                              Qt,
                              QT_VERSION_MAJOR,
                              QT_VERSION_MINOR,
                              QTimer,
                              Signal,
                              Slot)
from Nagstamon.qui.widgets.app import app
from Nagstamon.qui.widgets.labels import LabelAllOK
from Nagstamon.qui.widgets.server_vbox import ServerVBox
from Nagstamon.qui.widgets.statusbar import StatusBar
from Nagstamon.qui.widgets.toparea import TopArea
from Nagstamon.servers import (get_enabled_servers,
                               get_status_count,
                               servers)

# only on X11/Linux thirdparty path should be added because it contains the Xlib module
# needed to tell window manager via EWMH to keep Nagstamon window on all virtual desktops
if OS not in OS_NON_LINUX and not DESKTOP_WAYLAND:
    # extract thirdparty path from resources path - make submodules accessible by thirdparty modules
    sys.path.insert(0, sep.join(RESOURCES.split(sep)[0:-1] + ['thirdparty']))

    # Xlib for EWMH needs the file ~/.Xauthority and crashes if it does not exist
    xauthority_file = Path().home() / '.Xauthority'
    if not xauthority_file.exists():
        open(xauthority_file, 'a').close()

    from Nagstamon.thirdparty.ewmh import EWMH


class StatusWindow(QWidget):
    """
    Consists of statusbar, toparea and scrolling area.
    Either statusbar is shown or (toparea + scrolling area)
    """
    # sent by .resize_window()
    resizing = Signal()

    # send when windows opens, e.g. for stopping notifications
    showing = Signal()

    # send when window shrinks down to statusbar or closes
    hiding = Signal()

    # signal to be sent to all server workers to recheck all
    recheck = Signal()

    # signal to submit server to authentication dialog
    authenticate = Signal(str)

    # signal to be sent to all treeview workers to clear server event history
    # after 'Refresh'-button has been pressed
    clear_event_history = Signal()

    # signals for changing the state of the system tray icon
    systrayicon_enabled = Signal()
    systrayicon_disabled = Signal()

    # signal to request systray icon position for storing it in statuswindow_properties
    request_systrayicon_position = Signal()

    # shortcut to systray icon needed for connecting signals/slots
    injected_systrayicon = None

    # cached coordinates
    stored_x = 0
    stored_y = 0
    stored_height = 0
    stored_width = 0

    def __init__(self, dialogs=None, systrayicon=None):
        """
        Status window combined from status bar and popup window
        """
        QWidget.__init__(self)

        # immediately hide to avoid flicker on Windows and OSX
        self.hide()

        # avoid quitting when using Qt.Tool flag and closing settings dialog
        app.setQuitOnLastWindowClosed(False)

        # show tooltips even if popup window has no focus
        self.setAttribute(Qt.WidgetAttribute.WA_AlwaysShowToolTips)
        # title + icon
        self.setWindowTitle(AppInfo.NAME)
        self.setWindowIcon(QIcon(f'{RESOURCES}{sep}nagstamon.svg'))

        if OS == OS_MACOS:
            # avoid hiding window if it has no focus - necessary on OSX if using flag Qt.Tool
            self.setAttribute(Qt.WidgetAttribute.WA_MacAlwaysShowToolWindow)

        # ewmh.py in thirdparty directory needed to keep floating statusbar on all desktops in Linux
        if not OS in OS_NON_LINUX and not DESKTOP_WAYLAND:
            self.ewmh = EWMH()

        # set shortcut for dialogs icon to be used for connecting signals/slots
        self.injected_dialogs = dialogs
        # set shortcut for systray icon to be used for connecting signals/slots
        self.injected_systrayicon = systrayicon

        self.vbox = QVBoxLayout(self)  # global VBox
        self.vbox.setSpacing(0)  # no spacing
        self.vbox.setContentsMargins(0, 0, 0, 0)  # no margin

        self.statusbar = StatusBar(parent=self)  # statusbar HBox
        self.toparea = TopArea(parent=self)  # toparea HBox
        # no need to be seen first
        self.toparea.hide()

        self.servers_scrollarea = QScrollArea(self)  # scrollable area for server vboxes
        # avoid horizontal scrollbars
        self.servers_scrollarea.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        # necessary widget to contain vbox for servers
        self.servers_scrollarea_widget = QWidget(self.servers_scrollarea)
        self.servers_scrollarea.hide()

        self.vbox.addWidget(self.statusbar)
        self.vbox.addWidget(self.toparea)
        self.vbox.addWidget(self.servers_scrollarea)

        self.servers_vbox = QVBoxLayout(self.servers_scrollarea)  # VBox full of servers
        self.servers_vbox.setSpacing(0)
        self.servers_vbox.setContentsMargins(0, 0, 0, 0)

        self.label_all_ok = LabelAllOK(parent=self)
        self.label_all_ok.hide()
        self.servers_vbox.addWidget(self.label_all_ok)

        # test with OSX top menubar
        if OS == OS_MACOS:
            self.menubar = QMenuBar()
            action_exit = QAction('exit', self.menubar)
            action_settings = QAction('settings', self.menubar)
            self.menubar.addAction(action_settings)
            self.menubar.addAction(action_exit)

        # stored x y values for systemtray icon
        statuswindow_properties.icon_x = 0
        statuswindow_properties.icon_y = 0

        # if status_ok is true no server_vboxes are needed
        statuswindow_properties.status_ok = True

        # timer for waiting to set is_shown flag
        self.timer = QTimer(self)

        # react to open button in notification bubble
        dbus_connection.open_statuswindow.connect(self.show_window_from_notification_bubble)

        # connect logo of statusbar
        self.statusbar.logo.window_moved.connect(self.save_position)
        self.statusbar.logo.window_moved.connect(self.hide_window)
        self.statusbar.logo.window_moved.connect(self.correct_moving_position)
        self.statusbar.logo.window_moved.connect(self.update)
        self.statusbar.logo.mouse_pressed.connect(self.save_position)

        # after status summarization check if window has to be resized
        self.statusbar.resize.connect(self.adjust_size)

        # statusbar label has been entered by mouse -> show
        for label in self.statusbar.color_labels.values():
            label.mouse_entered.connect(self.show_window_after_checking_for_hover)
            label.mouse_released.connect(self.show_window_after_checking_for_clicking)
            label.mouse_released_in_window.connect(self.hide_window)

        # connect message label to hover
        self.statusbar.label_message.mouse_entered.connect(self.show_window_after_checking_for_hover)
        self.statusbar.label_message.mouse_released.connect(self.show_window_after_checking_for_clicking)
        self.statusbar.label_message.mouse_released_in_window.connect(self.hide_window)

        # when logo in toparea was pressed hurry up to save the position so the statusbar will not jump
        self.toparea.logo.window_moved.connect(self.save_position)
        self.toparea.logo.window_moved.connect(self.hide_window)
        self.toparea.logo.window_moved.connect(self.correct_moving_position)
        self.toparea.logo.window_moved.connect(self.update)
        self.toparea.logo.mouse_pressed.connect(self.save_position)
        self.toparea.logo.mouse_released_in_window.connect(self.hide_window)

        # when version label in toparea was pressed hurry up to save the position so the statusbar will not jump
        self.toparea.label_version.window_moved.connect(self.save_position)
        self.toparea.label_version.window_moved.connect(self.hide_window)
        self.toparea.label_version.window_moved.connect(self.correct_moving_position)
        self.toparea.label_version.window_moved.connect(self.update)
        self.toparea.label_version.mouse_pressed.connect(self.save_position)
        self.toparea.label_version.mouse_released_in_window.connect(self.hide_window)

        # when empty space in toparea was pressed hurry up to save the position so the statusbar will not jump
        self.toparea.label_empty_space.window_moved.connect(self.save_position)
        self.toparea.label_empty_space.window_moved.connect(self.hide_window)
        self.toparea.label_empty_space.window_moved.connect(self.correct_moving_position)
        self.toparea.label_empty_space.window_moved.connect(self.update)
        self.toparea.label_empty_space.mouse_pressed.connect(self.save_position)
        self.toparea.label_empty_space.mouse_released_in_window.connect(self.hide_window)

        # buttons in toparea
        self.toparea.button_recheck_all.clicked.connect(self.recheck_all)
        self.toparea.button_refresh.clicked.connect(self.refresh)
        self.toparea.button_settings.clicked.connect(self.hide_window)
        self.toparea.button_close.clicked.connect(self.hide_window)

        # if monitor was selected in combobox its monitor window is opened
        self.toparea.combobox_servers.monitor_opened.connect(self.hide_window)

        # due to lack of a better place the 'Delete cookies' button action is defined here
        self.injected_dialogs.weblogin.delete_web_cookies.connect(
            self.injected_dialogs.server.delete_web_cookies_action)
        self.injected_dialogs.server.delete_web_cookies.connect(self.injected_dialogs.server.delete_web_cookies_action)

        self.initialize()

    def initialize(self):
        """
        initialize the volatile widgets of the status window
        """
        # needed to store position of status window
        self.stored_y = 0
        self.stored_x = 0
        self.stored_width = 0
        self.stored_height = 0

        # worker and thread duo needed for notifications
        self.worker_notification_thread = QThread(parent=self)
        self.worker_notification = self.WorkerNotification(statuswindow_properties)

        # clean shutdown of thread
        self.worker_notification.finish.connect(self.finish_worker_notification_thread)

        # flashing statusbar
        self.worker_notification.start_flash.connect(self.statusbar.flash)
        self.worker_notification.stop_flash.connect(self.statusbar.reset)

        # desktop notification
        self.worker_notification.desktop_notification.connect(self.desktop_notification)

        # stop notification if window gets shown or hidden
        self.hiding.connect(self.worker_notification.stop)

        self.hiding.connect(self.move_while_hiding)

        self.worker_notification.moveToThread(self.worker_notification_thread)
        # start with low priority
        self.worker_notification_thread.start(QThread.Priority.LowestPriority)

        self.create_server_vboxes()

        # connect status window server vboxes to systray
        for server_vbox in self.servers_vbox.children():
            if 'server' in server_vbox.__dict__.keys():
                # tell systray after table was refreshed
                server_vbox.table.worker.new_status.connect(self.injected_systrayicon.show_state)
                # show error icon in systray
                server_vbox.table.worker.show_error.connect(self.injected_systrayicon.set_error)
                server_vbox.table.worker.hide_error.connect(self.injected_systrayicon.reset_error)

        self.servers_scrollarea_widget.setLayout(self.servers_vbox)
        self.servers_scrollarea.setWidget(self.servers_scrollarea_widget)
        self.servers_scrollarea.setWidgetResizable(True)

        # needed for moving the statuswindow
        statuswindow_properties.moving = False
        statuswindow_properties.relative_x = 0
        statuswindow_properties.relative_y = 0

        # helper values for QTimer.singleShot move attempt
        self.move_to_x = self.move_to_y = 0

        # flag to mark if window is shown or not
        if conf.windowed:
            statuswindow_properties.is_shown = True
        else:
            statuswindow_properties.is_shown = False

        # store show_window timestamp to avoid flickering window in KDE5 with systray
        statuswindow_properties.is_shown_timestamp = time()

        # store timestamp to avoid reappearing window shortly after clicking onto toparea
        statuswindow_properties.is_hiding_timestamp = time()

        # a thread + worker is necessary to do actions thread-safe in background
        # like debugging
        self.worker_thread = QThread(parent=self)
        self.worker = self.Worker()
        self.worker.moveToThread(self.worker_thread)
        # start thread and debugging loop if debugging is enabled
        if conf.debug_mode:
            self.worker_thread.started.connect(self.worker.debug_loop)
        # start with low priority
        self.worker_thread.start(QThread.Priority.LowestPriority)

        # clean shutdown of thread
        self.worker.finish.connect(self.finish_worker_thread)

        # finally show up
        self.set_mode()

    @Slot()
    def reinitialize(self):
        """
        delete volatile widgets and initialize them again
        """
        # stop both workers
        if hasattr(self, 'worker'):
            self.worker.running = False
            self.worker.finish.emit()
        if hasattr(self, 'worker_notification'):
            self.worker_notification.running = False
            self.worker_notification.finish.emit()
        # remove all server vboxes
        if hasattr(self, 'servers_vbox'):
            for vbox in self.servers_vbox.children():
                if hasattr(vbox, 'table') and hasattr(vbox.table, 'worker'):
                    vbox.table.worker.finish.emit()
                vbox.deleteLater()
        # initialize all volatile widgets
        self.initialize()

    def get_screen(self):
        """
        very hackish fix for https://github.com/HenriWahl/Nagstamon/issues/865
        should actually fit into qt.py but due to the reference to `app` it could only
        be solved here
        """
        # Qt6 has .screen() as replacement for QDesktopWidget...
        if QT_VERSION_MAJOR > 5:
            return self.screen()
        # ...and .screen() exists since Qt5 5.15...
        elif QT_VERSION_MINOR < 15:
            return app.desktop()
        # ...so newer ones can use .screen() again
        else:
            return self.screen()

    def set_mode(self):
        """
        apply presentation mode
        """
        # hide everything first
        self.hide()
        self.statusbar.hide()
        self.toparea.hide()
        self.servers_scrollarea.hide()

        # restore to normal window in case it was fullscreen
        self.showNormal()

        if conf.statusbar_floating:
            # show icon in dock if window is set
            if OS == OS_MACOS:
                # in floating mode always show dock icon - right now I am not able to
                # get the icon hidden
                hide_macos_dock_icon(False)

            # no need for systray
            self.systrayicon_disabled.emit()

            self.statusbar.show()

            # show statusbar/statuswindow on last saved position
            # when coordinates are inside known screens
            if get_screen_name(conf.position_x, conf.position_y):
                self.move(conf.position_x, conf.position_y)
            else:
                # get available desktop specs
                available_x = self.get_screen().availableGeometry().x()
                available_y = self.get_screen().availableGeometry().y()
                self.move(available_x, available_y)

            # statusbar and detail window should be frameless and stay on top
            # tool flag helps to be invisible in taskbar
            self.setWindowFlags(WINDOW_FLAGS)

            # show statusbar without being active, just floating
            self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)

            # necessary to be shown before Linux EWMH-mantra can be applied
            self.show()

            # X11/Linux needs some special treatment to get the statusbar floating on all virtual desktops
            if OS not in OS_NON_LINUX and not DESKTOP_WAYLAND:
                # get all windows...
                self.ewmh.setWmDesktop(self.winId().__int__(), 0xffffffff)
                self.ewmh.display.flush()

            # show statusbar/statuswindow on last saved position
            # when coordinates are inside known screens
            if get_screen_name(conf.position_x, conf.position_y):
                self.move(conf.position_x, conf.position_y)
            else:
                # get available desktop specs
                available_x = self.get_screen().availableGeometry().x()
                available_y = self.get_screen().availableGeometry().y()
                self.move(available_x, available_y)

            # need a close button
            self.toparea.button_close.show()

        elif conf.icon_in_systray:
            # no need for icon in dock if in systray
            if OS == OS_MACOS:
                hide_macos_dock_icon(conf.hide_macos_dock_icon)

            # statusbar and detail window should be frameless and stay on top
            # tool flag helps to be invisible in taskbar
            self.setWindowFlags(WINDOW_FLAGS)

            # show statusbar without being active, just floating
            self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)

            # yeah! systray!
            self.systrayicon_enabled.emit()

            # need a close button
            self.toparea.button_close.show()

            # reset systray icon position
            statuswindow_properties.icon_x = statuswindow_properties.icon_y = 0

        elif conf.fullscreen:
            # no need for systray
            self.systrayicon_disabled.emit()

            # needed permanently
            self.toparea.show()
            self.servers_scrollarea.show()

            # get screen geometry to get right screen to position window on
            screen_geometry = get_screen_geometry(conf.fullscreen_display)
            self.move(screen_geometry.x(), screen_geometry.y())

            # keep window entry in taskbar and thus no Qt.Tool
            self.setWindowFlags(Qt.WindowType.Widget | Qt.WindowType.FramelessWindowHint)

            # show statusbar actively
            self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating, False)

            # newer Qt5 seem to be better regarding fullscreen mode on non-OSX
            self.show_window()
            # fullscreen mode is rather buggy on everything other than OSX so just use a maximized window
            if OS == OS_MACOS:
                self.showFullScreen()
                # in fullscreen mode dock icon does not disturb because the dock is away anyway
                hide_macos_dock_icon(False)
            else:
                self.show()
                self.showMaximized()

            # no need for close button
            self.toparea.button_close.hide()

        elif conf.windowed:
            # show icon in dock if window is set
            if OS == OS_MACOS:
                # in windowed mode always show dock icon
                hide_macos_dock_icon(False)

            self.systrayicon_disabled.emit()

            # no need for close button
            self.toparea.button_close.hide()
            self.toparea.show()
            self.servers_scrollarea.show()

            # keep window entry in taskbar and thus no Qt.Tool
            self.setWindowFlags(Qt.WindowType.Widget)

            # show statusbar actively
            self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating, False)

            # some maybe sensible default
            self.setMinimumSize(700, 300)
            self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)

            # default maximum size
            self.setMaximumSize(16777215, 16777215)

            self.move(conf.position_x, conf.position_y)
            self.resize(conf.position_width, conf.position_height)

            # make sure window is shown
            self.show()
            self.showNormal()
            self.show_window()

            # make sure window comes up
            self.raise_()

        # force correct position of statuswindow
        self.adjust_size()

        # store position for showing/hiding statuswindow
        self.stored_x = self.x()
        self.stored_y = self.y()
        self.stored_width = self.width()

    def sort_server_vboxes(self):
        """
        sort ServerVBoxes alphabetically
        """
        # shortly after applying changes a QObject might hang around in the children list which should
        # be filtered out this way
        vboxes_dict = dict()
        for child in self.servers_vbox.children():
            if 'server' in child.__dict__.keys():
                vboxes_dict[child.server.name] = child

        # freshly set servers_scrollarea_widget and its layout servers_vbox
        servers_vbox_new = QVBoxLayout()  # VBox full of servers
        servers_vbox_new.setContentsMargins(0, 0, 0, 0)
        servers_vbox_new.setSpacing(0)

        # sort server vboxes
        for vbox in sorted(vboxes_dict):
            vboxes_dict[vbox].setParent(None)
            servers_vbox_new.addLayout(vboxes_dict[vbox])

        # add expanding stretching item at the end for fullscreen beauty
        servers_vbox_new.addSpacerItem(QSpacerItem(0, self.get_screen().availableGeometry().height(),
                                                   QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding))

        # switch to new servers_vbox
        self.servers_vbox = servers_vbox_new
        # necessary widget to contain vbox for servers
        self.servers_scrollarea_widget = QWidget()
        self.servers_scrollarea_widget.setLayout(self.servers_vbox)
        self.servers_scrollarea.setWidget(self.servers_scrollarea_widget)

        self.servers_scrollarea.contentsMargins().setTop(0)
        self.servers_scrollarea.contentsMargins().setBottom(0)

        del vboxes_dict

    @Slot(str)
    def create_server_vbox(self, name):
        """
        internally used to create enabled servers to be displayed
        """
        server = servers[name]

        # create server vboxed from current running servers
        if server.enabled:
            # display authentication dialog if password is not known
            if not conf.servers[server.name].save_password and \
                    not conf.servers[server.name].use_autologin and \
                    conf.servers[server.name].password == '' and \
                    not conf.servers[server.name].authentication == 'kerberos' and \
                    not conf.servers[server.name].authentication == 'web':
                self.authenticate.emit(server.name)

            # without parent, there is some flickering when starting
            server_vbox = ServerVBox(server, parent=self)

            # important to set correct server to worker, especially after server changes
            server_vbox.table.worker.server = server

            # connect to global resize signal
            server_vbox.table.ready_to_resize.connect(self.adjust_size)

            # tell statusbar to summarize after table was refreshed
            server_vbox.table.worker.new_status.connect(self.statusbar.summarize_states)
            server_vbox.table.worker.new_status.connect(self.raise_window_on_all_desktops)

            # if problems go themselves there is no need to notify user anymore
            server_vbox.table.worker.problems_vanished.connect(self.worker_notification.stop)

            # show the error message in statusbar
            server_vbox.table.worker.show_error.connect(self.statusbar.set_error)
            server_vbox.table.worker.hide_error.connect(self.statusbar.reset_error)

            # tell notification worker to do something AFTER the table was updated
            server_vbox.table.status_changed.connect(self.worker_notification.start)

            # and to update status window
            server_vbox.table.refreshed.connect(self.update_window)

            # hide statuswindow if authentication dialog is to be shown
            server_vbox.button_authenticate.clicked.connect(self.hide_window)

            # tell table it should remove freshness of formerly new items when window closes
            # because apparently the new events have been seen now
            self.hiding.connect(server_vbox.table.worker.unfresh_event_history)

            # stop notification if statuswindow pops up
            self.showing.connect(self.worker_notification.stop)

            # tell server worker to recheck all hosts and services
            self.recheck.connect(server_vbox.table.worker.recheck_all)

            # refresh table after changed settings
            self.injected_dialogs.settings.changed.connect(server_vbox.table.refresh)
            self.injected_dialogs.weblogin.page_loaded.connect(server_vbox.table.worker.get_status)

            # listen if statuswindow cries for event history clearance
            self.clear_event_history.connect(server_vbox.table.worker.unfresh_event_history)

            # statuswindow.servers_vbox.addLayout(statuswindow.create_server_vbox(servers[server_name]))
            self.servers_vbox.addLayout(server_vbox)
            self.sort_server_vboxes()

            return server_vbox
        else:
            return None

    def create_server_vboxes(self):
        """
        create VBox for each enabled server
        """
        for server in servers.values():
            if server.enabled:
                server_vbox = self.create_server_vbox(server.name)
                self.servers_vbox.addLayout(server_vbox)
        self.sort_server_vboxes()

    @Slot(str)
    def delete_server_vbox(self, name):
        """
        delete VBox for server with given name - called by signal from settings dialog
        """
        for vbox in self.servers_vbox.children():
            if vbox.server.name == name:
                # stop thread by falsificate running flag
                vbox.table.worker.running = False
                vbox.table.worker.finish.emit()
                break

    @Slot()
    def show_window_after_checking_for_clicking(self):
        """
        being called after clicking statusbar - check if window should be shown
        """
        if conf.popup_details_clicking:
            self.show_window()

    @Slot()
    def show_window_after_checking_for_hover(self):
        """
        being called after hovering over statusbar - check if window should be shown
        """
        if conf.popup_details_hover:
            self.show_window()

    @Slot()
    def show_window_from_notification_bubble(self):
        """
        show status window after button being clicked in notification bubble
        """
        if conf.statusbar_floating:
            self.show_window()
        elif conf.icon_in_systray:
            self.show_window_systrayicon()

    @Slot()
    def show_window_systrayicon(self):
        """
        handle clicks onto systray icon
        """
        if not statuswindow_properties.is_shown:
            # under unfortunate circumstances statusbar might have the the moving flag true
            # fix it here because it makes no sense but might cause non-appearing statuswindow‚
            statuswindow_properties.moving = False

            # hopefully no race condition here and the systray icon position will be retrieved just in time
            self.request_systrayicon_position.emit()

            # already show here because was closed before in hide_window()
            # best results achieved when doing .show() before .show_window()
            self.show()
            self.show_window()

            # Wayland workaround: Force an immediate update to prevent a black rectangle
            if DESKTOP_WAYLAND:
                self.update()
        else:
            self.hide_window()

    @Slot()
    def show_window_for_authentication(self):
        """
        show status window for authentication if not already shown
        especially needed for systray icon and floating mode
        """
        if not self.injected_dialogs.get_shown_dialogs():
            if conf.icon_in_systray:
                self.show_window_systrayicon()
            elif conf.statusbar_floating:
                self.show_window()

    @Slot()
    def show_window(self, event=None):
        """
        used to show status window when its appearance is triggered, also adjusts geometry
        """
        # do not show up when being dragged around
        if not statuswindow_properties.moving:
            # check if really all is OK
            for vbox in self.servers_vbox.children():
                if vbox.server.all_ok and \
                        vbox.server.status == '' and \
                        not vbox.server.refresh_authentication and \
                        not vbox.server.tls_error:
                    statuswindow_properties.status_ok = True
                else:
                    statuswindow_properties.status_ok = False
                    break

            # here we should check if scroll_area should be shown at all
            if not statuswindow_properties.status_ok:
                # store timestamp to avoid flickering as in https://github.com/HenriWahl/Nagstamon/issues/184
                statuswindow_properties.is_shown_timestamp = time()

                if not conf.fullscreen and not conf.windowed:
                    # attempt to avoid flickering on MacOSX - already hide statusbar here
                    self.statusbar.hide()
                    # show the other status window components
                    self.toparea.show()
                    self.servers_scrollarea.show()
                else:
                    self.label_all_ok.hide()

                for vbox in self.servers_vbox.children():
                    if not vbox.server.all_ok:
                        vbox.show_all()
                    # show at least server vbox header to notify about connection or other errors
                    elif vbox.server.status != '' or vbox.server.refresh_authentication or vbox.server.tls_error:
                        vbox.show_only_header()
                    elif vbox.server.all_ok and vbox.server.status == '':
                        vbox.hide_all()

                    # depending on authentication state show reauthentication button
                    if vbox.server.refresh_authentication:
                        vbox.button_authenticate.show()
                    else:
                        vbox.button_authenticate.hide()

                    # depending on TLS error show fix-TLS-button
                    if vbox.server.tls_error:
                        vbox.button_fix_tls_error.show()
                    else:
                        vbox.button_fix_tls_error.hide()

                if not conf.fullscreen and \
                        not conf.windowed:
                    # theory...
                    width, height, x, y = self.calculate_size()
                    # ...and practice
                    self.resize_window(width, height, x, y)
                    # switch on
                    if OS == OS_MACOS:
                        # delayed because of flickering window in OSX
                        self.timer.singleShot(200, self.set_shown)
                    else:
                        self.set_shown()

                    # avoid horizontally scrollable tables
                    self.adjust_dummy_columns()

                    self.show()

                    # Using the EWMH protocol to move the window to the active desktop.
                    # Seemed to be a problem on XFCE
                    # https://github.com/HenriWahl/Nagstamon/pull/199
                    if not OS in OS_NON_LINUX and conf.icon_in_systray:
                        try:
                            winid = self.winId().__int__()
                            deskid = self.ewmh.getCurrentDesktop()
                            self.ewmh.setWmDesktop(winid, deskid)
                            self.ewmh.display.flush()
                            # makes the window manager switch to the desktop where this widget has appeared
                            self.raise_()
                        except Exception:
                            # workaround for https://github.com/HenriWahl/Nagstamon/issues/246#issuecomment-220478066
                            pass

                    # tell others like notification that statuswindow shows up now
                    self.showing.emit()

            else:
                # hide vboxes in fullscreen and whole window in any other case if all is OK
                for vbox in self.servers_vbox.children():
                    vbox.hide_all()
                if conf.fullscreen or conf.windowed:
                    self.label_all_ok.show()
                if conf.icon_in_systray or conf.statusbar_floating:
                    self.hide_window()

            # If the mouse cursor drives too fast over and out the window will not be hidden.
            # Thus, we check again with this timer to catch missed mouse-outs.
            # causes trouble in Wayland so is disabled for it
            if conf.close_details_hover and \
                    conf.statusbar_floating and \
                    statuswindow_properties.is_shown and \
                    not DESKTOP_WAYLAND:
                self.periodically_check_window_under_mouse_and_hide()

    def periodically_check_window_under_mouse_and_hide(self):
        """
        Periodically check if window is under mouse and hide it if not
        """
        if not self.hide_window_if_not_under_mouse():
            self.timer.singleShot(1000, self.periodically_check_window_under_mouse_and_hide)

    def hide_window_if_not_under_mouse(self):
        """
        hide window if it is under mouse pointer
        """
        mouse_pos = QCursor.pos()
        # Check mouse cursor over window and an opened context menu or dropdown list
        if self.geometry().contains(mouse_pos.x(), mouse_pos.y()) or \
                not app.activePopupWidget() is None or \
                statuswindow_properties.is_shown:
            return False

        self.hide_window()
        return True

    @Slot()
    def update_window(self):
        """
        redraw window content, to be effective only when window is shown
        """
        if statuswindow_properties.is_shown or \
                conf.fullscreen or \
                (conf.windowed and statuswindow_properties.is_shown):
            self.show_window()

    @Slot()
    def hide_window(self):
        """
        hide window if not needed
        """
        if not conf.fullscreen and not conf.windowed:
            # only hide if shown and not locked or if not yet hidden if moving
            if statuswindow_properties.is_shown is True or \
                    statuswindow_properties.is_shown is True and \
                    statuswindow_properties.moving is True:
                # only hide if shown at least a fraction of a second
                # or has not been hidden a too short time ago
                if statuswindow_properties.is_shown_timestamp + 0.5 < time() or \
                        statuswindow_properties.is_hiding_timestamp + 0.1 < time():
                    if conf.statusbar_floating:
                        self.statusbar.show()
                    self.toparea.hide()
                    self.servers_scrollarea.hide()
                    # macOS needs this since Qt6 to avoid statuswindow size changeability
                    # looks silly but works to force using the own hint as hint
                    if OS == OS_MACOS:
                        self.setMinimumSize(self.sizeHint())
                        self.setMaximumSize(self.sizeHint())
                    else:
                        self.setMinimumSize(1, 1)
                    self.adjustSize()

                    if conf.icon_in_systray:
                        self.close()
                    else:
                        # show again after reorganizing widgets
                        self.show()

                    # switch off
                    statuswindow_properties.is_shown = False

                    # flag to reflect top-ness of window/statusbar
                    statuswindow_properties.top = False

                    # reset icon x y
                    statuswindow_properties.icon_x = 0
                    statuswindow_properties.icon_y = 0

                    if conf.windowed:
                        self.hide()

                    # store time of hiding
                    statuswindow_properties.is_hiding_timestamp = time()

                    # tell the world that window goes down
                    self.hiding.emit()

    @Slot()
    def move_while_hiding(self):
        """
        silly workaround to avoid flickering window in floating mode
        """
        # give some rest to avoid flickering
        sleep(0.0125)
        # move the window to the last stored position
        self.move(self.stored_x, self.stored_y)

    @Slot()
    def correct_moving_position(self):
        """
        correct position if moving and cursor started outside statusbar
        """
        if statuswindow_properties.moving:
            mouse_x = QCursor.pos().x()
            mouse_y = QCursor.pos().y()
            # when cursor is outside moved window correct the coordinates of statusbar/statuswindow
            if not self.geometry().contains(mouse_x, mouse_y):
                rect = self.geometry()
                corrected_x = int(mouse_x - rect.width() // 2)
                corrected_y = int(mouse_y - rect.height() // 2)
                # calculate new relative values
                statuswindow_properties.relative_x = mouse_x - corrected_x
                statuswindow_properties.relative_y = mouse_y - corrected_y
                self.move(corrected_x, corrected_y)
                del mouse_x, mouse_y, corrected_x, corrected_y

    def calculate_size(self):
        """
        get size of popup window
        """

        # only consider offset if it is configured
        if conf.systray_offset_use and conf.icon_in_systray:
            available_height = self.get_screen().availableGeometry().height() - conf.systray_offset
        else:
            available_height = self.get_screen().availableGeometry().height()

        available_width = self.get_screen().availableGeometry().width()
        available_x = self.get_screen().availableGeometry().x()
        available_y = self.get_screen().availableGeometry().y()

        # Workaround for Cinnamon + GNOME Flashback
        if OS not in OS_NON_LINUX and conf.enable_position_fix:
            if available_x == 0:
                available_x = available_width
            if available_y == 0:
                available_y = available_height

        # take whole screen height into account when deciding about upper/lower-ness
        # add available_y because it might vary on differently setup screens
        # calculate top-ness only if window is closed
        if conf.statusbar_floating:
            if self.y() < self.get_screen().geometry().height() // 2 + available_y:
                statuswindow_properties.top = True
            else:
                statuswindow_properties.top = False

            # always take the stored position of the statusbar
            x = self.stored_x

        elif conf.icon_in_systray or conf.windowed:
            # Wayland fallback for absolute coordinates
            if DESKTOP_WAYLAND and statuswindow_properties.icon_x == 0:
                statuswindow_properties.icon_x = available_x + available_width - 50
                statuswindow_properties.icon_y = available_y + available_height

            # Determine top/bottom placement AFTER retrieving/setting coordinates
            if statuswindow_properties.icon_y < self.get_screen().geometry().height() // 2 + available_y:
                statuswindow_properties.top = True
            else:
                statuswindow_properties.top = False

            # just a little buffer to let systrayicon.retrieve_icon_position() time to do its work, just in case
            while statuswindow_properties.icon_x == 0:
                sleep(0.01)

            x = statuswindow_properties.icon_x

        # get height from table widgets
        real_height = self.get_real_height()

        # width simply will be the current screen maximal width - less hassle!
        if self.get_real_width() > available_width:
            width = available_width
            x = available_x
        else:
            width = self.get_real_width()

            if width < self.toparea.sizeHint().width():
                width = self.toparea.sizeHint().width()

            # always take the stored width of the statusbar into account
            x -= int(width // 2) + int(self.stored_width // 2)

            # check left and right limits of x
            if x < available_x:
                x = available_x
            if x + width > available_x + available_width:
                x = available_x + available_width - width

        if conf.statusbar_floating:
            # when statusbar resides in the uppermost part of current screen extend from top to bottom
            if statuswindow_properties.top:
                y = self.y()
                if self.y() + real_height < available_height + available_y:
                    height = real_height
                else:
                    height = available_height - self.y() + available_y

            # when statusbar hangs around in lowermost part of current screen extend from bottom to top
            else:
                # when height is too large for current screen cut it
                if self.y() + self.height() - real_height < available_y:
                    height = self.get_screen().geometry().height() - available_y - (
                            self.get_screen().geometry().height() - (self.y() + self.height()))
                    y = available_y
                else:
                    height = real_height
                    y = self.y() + self.height() - height

        elif conf.icon_in_systray or conf.windowed:
            # when systrayicon resides in the uppermost part of current screen extend from top to bottom
            if statuswindow_properties.top:
                # when being top y is of course the available one
                y = available_y
                if self.y() + real_height < available_height + available_y:
                    height = real_height
                else:
                    # if bigger than screen shrink to maximal real_height
                    height = available_height - available_y
            # when statusbar hangs around in lowermost part of current screen extend from bottom to top
            else:
                if available_height < real_height:
                    y = available_y
                    height = available_height
                else:
                    y = available_height - real_height
                    height = real_height

        return width, height, x, y

    def resize_window(self, width, height, x, y):
        """
        resize status window according to its new size
        """
        # store position for restoring it when hiding - only if not shown of course
        if statuswindow_properties.is_shown is False:
            self.stored_x = self.x()
            self.stored_y = self.y()
            self.stored_width = self.width()
            self.stored_height = self.height()

        if OS == OS_WINDOWS:
            # absolutely strange, but no other solution available
            # - Only on Windows the statusbar is moving FIRST before resizing - no matter which
            #   order was used
            # - Dirty workaround:
            #   - store x and y in .move_to_*
            #   - start helper move_timer by timer singleshot to give statusbar some time to hide
            self.move_to_x, self.move_to_y = x, y
            self.timer.singleShot(10, self.move_timer)
        else:
            self.move(x, y)

        self.setMaximumSize(width, height)
        self.setMinimumSize(width, height)

        self.adjustSize()

        return True

    @Slot()
    def move_timer(self):
        """
        helper for move by QTimer.singleShot - attempt to avoid flickering on Windows
        """
        self.move(self.move_to_x, self.move_to_y)

    @Slot()
    def adjust_size(self):
        """
        resize window if shown and needed
        """
        # avoid race condition when waiting for password dialog
        if 'is_shown' in statuswindow_properties.__dict__:
            if not conf.fullscreen and not conf.windowed:
                # fully displayed statuswindow
                if statuswindow_properties.is_shown is True:
                    width, height, x, y = self.calculate_size()
                    self.adjust_dummy_columns()
                else:
                    # statusbar only
                    hint = self.sizeHint()
                    # on MacOSX and Windows statusbar will not shrink automatically, so this workaround hopefully helps
                    width = hint.width()
                    height = hint.height()
                    x = self.x()
                    y = self.y()
                    self.setMaximumSize(hint)
                    self.setMinimumSize(hint)
                    del hint
                self.resize_window(width, height, x, y)

                del width, height, x, y
            else:
                self.adjust_dummy_columns()

    @Slot()
    def adjust_dummy_columns(self):
        """
        calculate the widest width of all server tables to hide dummy column at the widest one
        """
        max_width = 0
        max_width_table = None
        for server in self.servers_vbox.children():
            # if table is wider than current max_width take its width as max_width
            if server.table.get_real_width() > max_width:
                max_width = server.table.get_real_width()
                max_width_table = server.table

        # widest table does not need the dummy column #9
        for server in self.servers_vbox.children():
            if max_width_table == server.table and max_width == server.table.width():
                # hide dummy column as here is the most stretched table
                server.table.setColumnHidden(9, True)
                server.table.header().setStretchLastSection(False)
            else:
                # show dummy column because some other table is wider
                server.table.setColumnHidden(9, False)
                server.table.header().setStretchLastSection(True)
        del max_width, max_width_table
        return True

    @Slot()
    def save_position(self):
        """
        store position for restoring it when hiding
        """
        if not statuswindow_properties.is_shown:
            self.stored_x = self.x()
            self.stored_y = self.y()
            self.stored_width = self.width()
            self.stored_height = self.height()

    def leaveEvent(self, event):
        """
        check if popup has to be hidden depending on mouse position
        """
        # depending on display mode the leave time offset shall be different because
        # it may be too short in systray mode and lead to flickering window
        if conf.statusbar_floating:
            leave_time_offset = 0.25
        elif conf.icon_in_systray:
            # offset is max 1 and smaller if window is smaller too
            leave_time_offset = self.height() / self.get_screen().availableGeometry().height()
        else:
            leave_time_offset = 0

        # check first if popup has to be shown by hovering or clicking
        if conf.close_details_hover and \
                not conf.fullscreen and \
                not conf.windowed and \
                statuswindow_properties.is_shown_timestamp + leave_time_offset < time():
            # only hide window if cursor is outside of it
            mouse_x = QCursor.pos().x()
            mouse_y = QCursor.pos().y()
            # <= and >= necessary because sometimes mouse_* is the same as self.*()
            if mouse_x <= self.x() or mouse_x >= self.x() + self.width() or \
                    mouse_y <= self.y() or mouse_y >= self.y() + self.height():
                self.hide_window()

    def closeEvent(self, event):
        """
        window close
        """
        # check first if popup has to be shown by hovering or clicking
        if conf.windowed:
            exit()

    def get_real_width(self):
        """
        calculate the widest width of all server tables
        """
        width = 0
        for server in self.servers_vbox.children():
            # if table is wider than window adjust with to table
            if server.table.isVisible() and server.table.get_real_width() > width:
                width = server.table.get_real_width()
            # if header in server vbox is wider than width adjust the latter
            if server.header.sizeHint().width() > width:
                width = server.header.sizeHint().width()
        return width

    def get_real_height(self):
        """
        calculate summary of all heights of all server tables plus height of toparea
        """
        height = 0
        for vbox in self.servers_vbox.children():
            height += vbox.get_real_height()

        # add size of toparea and 2 times the MARGIN (top and bottom)
        height += self.toparea.sizeHint().height() + 2

        return height

    def set_shown(self):
        """
        might help to avoid flickering on MacOSX, in cooperation with QTimer
        """
        statuswindow_properties.is_shown = True

    @Slot()
    def save_position_to_conf(self):
        """
        store position of statuswindow/statusbar
        """
        # only useful if statusbar is floating
        if conf.statusbar_floating:
            # minimize window to statusbar only to get real position
            self.hide_window()
            conf.position_x = self.x()
            conf.position_y = self.y()
        if conf.windowed:
            conf.position_x = self.x()
            conf.position_y = self.y()
            conf.position_width = self.width()
            conf.position_height = self.height()
        # store position of statuswindow/statusbar
        conf.save_config()

    @Slot(str, str)
    def show_message(self, msg_type, message):
        """
        show message from other thread like MediaPlayer
        """
        title = " ".join((AppInfo.NAME, msg_type))
        if msg_type == 'warning':
            return QMessageBox.warning(self, title, message)

        elif msg_type == 'information':
            return QMessageBox.information(self, title, message)
        return None

    @Slot()
    def recheck_all(self):
        """
        tell servers to recheck all hosts and services
        """
        self.recheck.emit()

    @Slot()
    def refresh(self):
        """
        tell all enabled servers to refresh their information
        """
        # unfresh event history of servers
        self.clear_event_history.emit()

        for server in get_enabled_servers():
            if conf.debug_mode:
                server.debug(server=server.name, debug='Refreshing all hosts and services')

            # manipulate server thread counter so get_status loop will refresh when next looking
            # at thread counter
            server.thread_counter = conf.update_interval_seconds + 1

    @Slot(dict)
    def desktop_notification(self, current_status_count):
        """
        show desktop notification - must be called from same thread as DBus intialization
        """
        # compile message from status counts
        message = ''
        for state in ['DOWN', 'UNREACHABLE', 'DISASTER', 'CRITICAL', 'HIGH', 'AVERAGE', 'WARNING', 'INFORMATION',
                      'UNKNOWN']:
            if current_status_count[state] > 0:
                message += '{0} {1} '.format(str(current_status_count[state]), state)

        if not message == '':
            # due to mysterious DBus-Crashes
            # see https://github.com/HenriWahl/Nagstamon/issues/320
            try:
                dbus_connection.show(AppInfo.NAME, message)
            except Exception:
                print_exc(file=stdout)

    @Slot()
    def raise_window_on_all_desktops(self):
        """
        experimental workaround for floating-statusbar-only-on-one-virtual-desktop-after-a-while bug
        see https://github.com/HenriWahl/Nagstamon/issues/217
        """
        if conf.windowed:
            return
        # X11/Linux needs some special treatment to get the statusbar floating on all virtual desktops
        if OS not in OS_NON_LINUX and not DESKTOP_WAYLAND:
            # get all windows...
            winid = self.winId().__int__()
            self.ewmh.setWmDesktop(winid, 0xffffffff)
            self.ewmh.display.flush()

        # apparently sometime the floating statusbar vanishes in the background
        # lets try here to keep it on top - only if not fullscreen
        if not conf.fullscreen and not conf.windowed and not OS == OS_WINDOWS:
            self.setWindowFlags(WINDOW_FLAGS)

        # again and again try to keep that statuswindow on top!
        if OS == OS_WINDOWS and \
                not conf.fullscreen and \
                not conf.windowed and \
                app.activePopupWidget() is None:
            try:
                self.raise_()
            except Exception as error:
                # apparently a race condition could occur on set_mode() - grab it here and continue
                print(error)

    @Slot()
    def finish_worker_thread(self):
        """
        attempt to shut down thread cleanly
        """
        # stop debugging
        statuswindow_properties.worker_debug_loop_looping = False
        # tell thread to quit
        self.worker_thread.quit()
        # wait until thread is really stopped
        self.worker_thread.wait()

    @Slot()
    def finish_worker_notification_thread(self):
        """
        attempt to shut down thread cleanly
        """
        # tell thread to quit
        self.worker_notification_thread.quit()
        # wait until thread is really stopped
        self.worker_notification_thread.wait()

    @Slot(str)
    def remove_previous_server_vbox(self, previous_server_name):
        # remove old server vbox from the status window if still running
        for vbox in self.servers_vbox.children():
            if vbox.server.name == previous_server_name:
                # disable server
                vbox.server.enabled = False
                # stop thread by falsificate running flag
                vbox.table.worker.running = False
                vbox.table.worker.finish.emit()
                # nothing more to do
                break

    @Slot()
    def decrease_shown_timestamp(self):
        """
        small workaround for the timestamp trick to avoid flickering
        if the 'Settings' button was clicked too fast the timestamp difference
        is too short and the statuswindow will keep open
        modifying the timestamp could help
        """
        statuswindow_properties.is_shown_timestamp -= 1

    @Slot()
    def exit(self):
        """
        stop all child threads before quitting instance
        """
        # store position of statuswindow/statusbar
        self.save_position_to_conf()

        # save configuration
        conf.save_config()

        # hide statuswindow first to avoid lag when waiting for finished threads
        self.hide()

        # stop statuswindow workers
        self.worker.finish.emit()
        self.worker_notification.finish.emit()

        # tell all treeview threads to stop
        for server_vbox in self.servers_vbox.children():
            server_vbox.table.worker.finish.emit()

        app.exit()

    class Worker(QObject):
        """
        run a thread, for example, for debugging
        """
        # send signal if ready to stop
        finish = Signal()

        def __init__(self):
            QObject.__init__(self)
            # flag to decide if the thread has to run or to be stopped
            self.running = True
            # flag if debug_loop is looping
            self.debug_loop_looping = False
            # default debug dile does not exist
            self.debug_file = None

        def open_debug_file(self):
            # open file and truncate
            self.debug_file = open(conf.debug_file, "w")

        def close_debug_file(self):
            # close and reset file
            self.debug_file.close()
            self.debug_file = None

        @Slot()
        def debug_loop(self):
            """
            if debugging is enabled, poll debug_queue list and print/write its contents
            """
            if conf.debug_mode:
                statuswindow_properties.worker_debug_loop_looping = True

                # as long thread is supposed to run
                while self.running and statuswindow_properties.worker_debug_loop_looping:
                    # only log something if there is something to tell
                    while len(debug_queue) > 0:
                        # always get the oldest item of queue list - FIFO
                        debug_line = (debug_queue.pop(0))
                        # output to console
                        print(debug_line)
                        if conf.debug_to_file:
                            # if there is no file handle available get it
                            if self.debug_file is None:
                                self.open_debug_file()
                            # log line per line
                            self.debug_file.write(debug_line + "\n")
                    # wait second until the next poll
                    sleep(1)

                # unset looping
                statuswindow_properties.worker_debug_loop_looping = False
                # close file if any
                if self.debug_file is not None:
                    self.close_debug_file()
            else:
                # set the flag to tell debug loop it should stop, please
                self.debug_loop_looping = False

    class WorkerNotification(QObject):
        """
        run a thread for doing all notification stuff
        """
        # tell statusbar labels to flash
        start_flash = Signal()
        stop_flash = Signal()

        # tell mediaplayer to load and play sound file
        load_sound = Signal(str)
        play_sound = Signal()

        # tell statuswindow to use desktop notification
        desktop_notification = Signal(dict)

        # only one enabled server should have the right to send play_sound signal
        notifying_server = ''

        # desktop notification needs to store count of states
        status_count = dict()

        # send signal if ready to stop
        finish = Signal()

        def __init__(self, statuswindow_properties=None):
            QObject.__init__(self)
            self.statuswindow_properties = statuswindow_properties

        @Slot(str, str, str)
        def start(self, server_name, worst_status_diff, worst_status_current):
            """
            start notification
            """
            if conf.notification:
                # only if not notifying yet or the current state is worse than the prior AND
                # only when the current state is configured to be honking about
                if (STATES.index(worst_status_diff) > STATES.index(
                        self.statuswindow_properties.worst_notification_status) or
                    self.statuswindow_properties.is_notifying is False) and \
                        conf.__dict__[f'notify_if_{worst_status_diff.lower()}'] is True:
                    # keep last worst state worth a notification for comparison 3 lines above
                    self.statuswindow_properties.worst_notification_status = worst_status_diff
                    # set flag to avoid innecessary notification
                    self.statuswindow_properties.is_notifying = True
                    if self.statuswindow_properties.notifying_server == '':
                        statuswindow_properties.notifying_server = server_name

                    # flashing statusbar
                    if conf.notification_flashing:
                        self.start_flash.emit()

                    # Play default sounds via mediaplayer
                    if conf.notification_sound:
                        sound_file = ''
                        # at the moment there are only sounds for down, critical and warning
                        # only honk if notifications are wanted for this state
                        if worst_status_diff in STATES_SOUND:
                            if conf.notification_default_sound:
                                # default .wav sound files are in resources folder
                                sound_file = '{0}{1}{2}.wav'.format(RESOURCES, sep, worst_status_diff.lower())
                            elif conf.notification_custom_sound:
                                sound_file = conf.__dict__[
                                    'notification_custom_sound_{0}'.format(worst_status_diff.lower())]

                            # only one enabled server should access the mediaplayer
                            if statuswindow_properties.notifying_server == server_name:
                                # once loaded file will be played by every server, even if it is
                                # not the statuswindow_properties.notifying_server that loaded it
                                self.load_sound.emit(sound_file)
                                self.play_sound.emit()

                    # Notification actions
                    if conf.notification_actions:
                        if conf.notification_action_warning is True and worst_status_diff == 'WARNING':
                            self.execute_action(server_name, conf.notification_action_warning_string)
                        if conf.notification_action_critical is True and worst_status_diff == 'CRITICAL':
                            self.execute_action(server_name, conf.notification_action_critical_string)
                        if conf.notification_action_down is True and worst_status_diff == 'DOWN':
                            self.execute_action(server_name, conf.notification_action_down_string)

                # Notification action OK
                if worst_status_current == 'UP' and \
                        conf.notification_actions and conf.notification_action_ok:
                    self.execute_action(server_name, conf.notification_action_ok_string)

                # Custom event notification - valid vor ALL events, thus without status comparison
                if conf.notification_actions is True and conf.notification_custom_action is True:
                    # temporarily used to collect executed events
                    events_list = []
                    events_string = ''

                    # if no single notifications should be used (default) put all events into one string, separated by separator
                    if conf.notification_custom_action_single is False:
                        for server in get_enabled_servers():
                            # list comprehension only considers events which are new, ergo True
                            events_list += [k for k, v in
                                            server.events_notification.items() if v is True]

                        # create string for no-single-event-notification of events separated by separator
                        events_string = conf.notification_custom_action_separator.join(events_list)

                        # clear already notified events setting them to False
                        for server in get_enabled_servers():
                            for event in [k for k, v in
                                          server.events_notification.items() if v is True]:
                                server.events_notification[event] = False
                    else:
                        for server in get_enabled_servers():
                            for event in [k for k, v in server.events_notification.items() if v is True]:
                                custom_action_string = conf.notification_custom_action_string.replace('$EVENT$',
                                                                                                      '$EVENTS$')
                                custom_action_string = custom_action_string.replace('$EVENTS$', event)
                                # execute action
                                self.execute_action(server_name, custom_action_string)
                                # clear already notified events setting them to False
                                server.events_notification[event] = False

                    # if events got filled display them now
                    if events_string != '':
                        # in case a single action per event has to be executed
                        custom_action_string = conf.notification_custom_action_string.replace('$EVENT$', '$EVENTS$')
                        # insert real event(s)
                        custom_action_string = custom_action_string.replace('$EVENTS$', events_string)
                        # execute action
                        self.execute_action(server_name, custom_action_string)
                else:
                    # set all events to False to ignore them in the future
                    for event in servers[server_name].events_notification:
                        servers[server_name].events_notification[event] = False

                # repeated sound
                # only let one enabled server play sound to avoid a larger cacophony
                if statuswindow_properties.is_notifying and \
                        conf.notification_sound_repeat and \
                        statuswindow_properties == server_name:
                    self.play_sound.emit()

                # desktop notification
                if conf.notification_desktop:
                    # get status count from servers
                    current_status_count = get_status_count()
                    if current_status_count != self.status_count:
                        self.desktop_notification.emit(current_status_count)
                    # store status count for next comparison
                    self.status_count = current_status_count
                    del current_status_count

        @Slot()
        def stop(self):
            """
            stop notification if there is no need anymore
            """
            if self.statuswindow_properties.is_notifying:
                self.statuswindow_properties.worst_notification_status = 'UP'
                self.statuswindow_properties.is_notifying = False

                # no more flashing statusbar and systray
                self.stop_flash.emit()

                # reset notifying server, waiting for next notification
                statuswindow_properties.notifying_server = ''

        def execute_action(self, server_name, custom_action_string):
            """
            execute custom action
            """
            if conf.debug_mode:
                servers[server_name].debug(debug='NOTIFICATION: ' + custom_action_string)
            Popen(custom_action_string, shell=True)



XK_leftcaret = 0xba3
XK_rightcaret = 0xba6
XK_downcaret = 0xba8
XK_upcaret = 0xba9
XK_overbar = 0xbc0
XK_downtack = 0xbc2
XK_upshoe = 0xbc3
XK_downstile = 0xbc4
XK_underbar = 0xbc6
XK_jot = 0xbca
XK_quad = 0xbcc
XK_uptack = 0xbce
XK_circle = 0xbcf
XK_upstile = 0xbd3
XK_downshoe = 0xbd6
XK_rightshoe = 0xbd8
XK_leftshoe = 0xbda
XK_lefttack = 0xbdc
XK_righttack = 0xbfc



# Xlib.protocol.structs -- some common request structures
#
#    Copyright (C) 2000 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,  USA

# Xlib modules
from Xlib import X

# Xlib.protocol modules
from Xlib.protocol import rq

def WindowValues(arg):
    return rq.ValueList( arg, 4, 0,
                         rq.Pixmap('background_pixmap'),
                         rq.Card32('background_pixel'),
                         rq.Pixmap('border_pixmap'),
                         rq.Card32('border_pixel'),
                         rq.Gravity('bit_gravity'),
                         rq.Gravity('win_gravity'),
                         rq.Set('backing_store', 1,
                                (X.NotUseful, X.WhenMapped, X.Always)),
                         rq.Card32('backing_planes'),
                         rq.Card32('backing_pixel'),
                         rq.Bool('override_redirect'),
                         rq.Bool('save_under'),
                         rq.Card32('event_mask'),
                         rq.Card32('do_not_propagate_mask'),
                         rq.Colormap('colormap'),
                         rq.Cursor('cursor'),
                         )

def GCValues(arg):
    return rq.ValueList( arg, 4, 0,
                         rq.Set('function', 1,
                                (X.GXclear, X.GXand, X.GXandReverse,
                                 X.GXcopy, X.GXandInverted, X.GXnoop,
                                 X.GXxor, X.GXor, X.GXnor, X.GXequiv,
                                 X.GXinvert, X.GXorReverse, X.GXcopyInverted,
                                 X.GXorInverted, X.GXnand, X.GXset)),
                         rq.Card32('plane_mask'),
                         rq.Card32('foreground'),
                         rq.Card32('background'),
                         rq.Card16('line_width'),
                         rq.Set('line_style', 1,
                                (X.LineSolid, X.LineOnOffDash, X.LineDoubleDash)),
                         rq.Set('cap_style', 1,
                                (X.CapNotLast, X.CapButt,
                                 X.CapRound, X.CapProjecting)),
                         rq.Set('join_style', 1,
                                (X.JoinMiter, X.JoinRound, X.JoinBevel)),
                         rq.Set('fill_style', 1,
                                (X.FillSolid, X.FillTiled,
                                 X.FillStippled, X.FillOpaqueStippled)),
                         rq.Set('fill_rule', 1,
                                (X.EvenOddRule, X.WindingRule)),
                         rq.Pixmap('tile'),
                         rq.Pixmap('stipple'),
                         rq.Int16('tile_stipple_x_origin'),
                         rq.Int16('tile_stipple_y_origin'),
                         rq.Font('font'),
                         rq.Set('subwindow_mode', 1,
                                (X.ClipByChildren, X.IncludeInferiors)),
                         rq.Bool('graphics_exposures'),
                         rq.Int16('clip_x_origin'),
                         rq.Int16('clip_y_origin'),
                         rq.Pixmap('clip_mask'),
                         rq.Card16('dash_offset'),
                         rq.Card8('dashes'),
                         rq.Set('arc_mode', 1, (X.ArcChord, X.ArcPieSlice))
                         )



TimeCoord = rq.Struct(
    rq.Card32('time'),
    rq.Int16('x'),
    rq.Int16('y'),
    )

Host = rq.Struct(
    rq.Set('family', 1, (X.FamilyInternet, X.FamilyDECnet, X.FamilyChaos)),
    rq.Pad(1),
    rq.LengthOf('name', 2),
    rq.List('name', rq.Card8Obj)
    )

CharInfo = rq.Struct(
    rq.Int16('left_side_bearing'),
    rq.Int16('right_side_bearing'),
    rq.Int16('character_width'),
    rq.Int16('ascent'),
    rq.Int16('descent'),
    rq.Card16('attributes'),
    )

FontProp = rq.Struct(
    rq.Card32('name'),
    rq.Card32('value'),
    )

ColorItem = rq.Struct(
    rq.Card32('pixel'),
    rq.Card16('red'),
    rq.Card16('green'),
    rq.Card16('blue'),
    rq.Card8('flags'),
    rq.Pad(1),
    )


RGB = rq.Struct(
    rq.Card16('red'),
    rq.Card16('green'),
    rq.Card16('blue'),
    rq.Pad(2),
    )


Point = rq.Struct(
    rq.Int16('x'),
    rq.Int16('y'),
    )

Segment = rq.Struct(
    rq.Int16('x1'),
    rq.Int16('y1'),
    rq.Int16('x2'),
    rq.Int16('y2'),
    )

Rectangle = rq.Struct(
    rq.Int16('x'),
    rq.Int16('y'),
    rq.Card16('width'),
    rq.Card16('height'),
   )

Arc = rq.Struct(
    rq.Int16('x'),
    rq.Int16('y'),
    rq.Card16('width'),
    rq.Card16('height'),
    rq.Int16('angle1'),
    rq.Int16('angle2'),
   )



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from Nagstamon.config import conf
from Nagstamon.qui.constants import HEADERS_HEADERS
from Nagstamon.qui.globals import font_icons
from Nagstamon.qui.qt import (QAbstractTableModel,
                              Qt,
                              QVariant,
                              Signal,
                              Slot)


class Model(QAbstractTableModel):
    """
    model for storing status data to be presented in Treeview-table
    """

    model_data_array_filled = Signal()

    # list of lists for storage of status data
    data_array = list()

    # cache row and column count
    row_count = 0
    column_count = len(HEADERS_HEADERS)

    # tell treeview if flags columns should be hidden or not
    hosts_flags_column_needed = Signal(bool)
    services_flags_column_needed = Signal(bool)

    def __init__(self, server, parent=None):
        QAbstractTableModel.__init__(self, parent=parent)
        self.server = server

    def rowCount(self, parent):
        """
        overridden method to get number of rows
        """
        return self.row_count

    def columnCount(self, parent):
        """
        overridden method to get number of columns
        """
        return self.column_count

    def headerData(self, column, orientation, role):
        """
        overridden method to get headers of columns
        """
        if role == Qt.ItemDataRole.DisplayRole:
            return HEADERS_HEADERS[column]
        return None

    @Slot(list, dict)
    # @Slot(list)
    def fill_data_array(self, data_array, info):
        """
        fill data_array for model
        """
        # tell treeview that model is about to change - necessary because
        # otherwise new number of rows would not be applied
        self.beginResetModel()

        # first empty the data storage
        del self.data_array[:]

        # use delivered data array
        self.data_array = data_array

        # cache row_count
        self.row_count = len(self.data_array)

        # tell treeview if flags columns are needed
        self.hosts_flags_column_needed.emit(info['hosts_flags_column_needed'])
        self.services_flags_column_needed.emit(info['services_flags_column_needed'])

        # new model applied
        self.endResetModel()

        self.model_data_array_filled.emit()

    def data(self, index, role):
        """
        overridden method for data delivery for treeview
        """
        if role == Qt.ItemDataRole.DisplayRole:
            return self.data_array[index.row()][index.column()]

        elif role == Qt.ItemDataRole.ForegroundRole:
            return self.data_array[index.row()][10]

        elif role == Qt.ItemDataRole.BackgroundRole:
            return self.data_array[index.row()][11]

        elif role == Qt.ItemDataRole.FontRole:
            if index.column() == 1:
                return font_icons
            elif index.column() == 3:
                return font_icons
            else:
                return QVariant
        # provide icons via Qt.UserRole
        elif role == Qt.ItemDataRole.UserRole:
            # depending on host or service column return host or service icon list
            return self.data_array[index.row()][7 + index.column()]

        elif role == Qt.ItemDataRole.ToolTipRole:
            # only if tooltips are wanted show them, combining host + service + status_info
            if conf.show_tooltips:
                return (f'<div style=white-space:pre;margin:3px;>'
                        f'<b>{self.data_array[index.row()][0]}: {self.data_array[index.row()][2]}</b>'
                        f'</div>'
                        f'{self.data_array[index.row()][8]}')
            else:
                return QVariant
        return None


# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import json

from . import CentreonAPI
from ..Generic import GenericServer
from Nagstamon.config import conf

class CentreonServer(GenericServer):
    """
    Use this class as switch
    """
    TYPE = 'Centreon'
    def __init__(self, **kwds):
        # like all servers we need to initialize Generic
        GenericServer.__init__(self, **kwds)
        # due to not being initialized right now we need to access config directly to get this instance's config
        server_conf = conf.servers.get(kwds.get('name'))
        if server_conf.enabled:
            self.authentication = conf.servers[self.get_name()].authentication
            # because auf being very early in init process the property ignore_cert is not known yet
            # add it here to be able to fetch URL and ignore certs if activated
            self.ignore_cert = conf.servers[self.get_name()].ignore_cert
            self.custom_cert_use = conf.servers[self.get_name()].custom_cert_use
            # This URL exists at least from Centreon 22.x - if not accessible it must be legacy
            versions_raw = self.fetch_url(f'{conf.servers[self.get_name()].monitor_cgi_url}/api/latest/platform/versions', no_auth=True, giveback='raw')
            self.debug(server='[' + self.get_name() + ']', debug='Page retrieval to detect the version, status code : %s' % (str(versions_raw.status_code)))
            if versions_raw.status_code == 200:
                try:
                    data = json.loads(versions_raw.result)
                except Exception as e:
                    self.debug(server='[' + self.get_name() + ']', debug='Page retrieval to detect the version, ERROR when decoding JSON')
                    self.enable = False
                    self.isChecking = False
                    return None

                ver_major = int(data["web"]["major"])
                ver_minor = int(data["web"]["minor"])
                # API V2 is usable only after 21.04 (not tested), ressources endpoint is buggy in 20.10
                if ver_major >= 21:
                    self.debug(server='[' + self.get_name() + ']', debug='Loading class API, Centreon version : ' + str(ver_major) + '.' + str(ver_minor))
                    from .CentreonAPI import CentreonServer as CentreonServerReal
                elif ver_major < 21:
                    # Fallback to Legacy
                    self.debug(server='[' + self.get_name() + ']', debug='Loading class LEGACY, Centreon version : ' + str(ver_major) + '.' + str(ver_minor))
                    from .CentreonLegacy import CentreonServer as CentreonServerReal
                else:
                    self.debug(server='[' + self.get_name() + ']', debug='Something wrong happens with Centreon, version detected : ' + str(ver_major) + '.' + str(ver_minor))
                    self.enable = False
                    self.isChecking = False
                    return None
            else:
                # Try to check if only it’s a Centreon
                versions_raw = self.fetch_url(f'{server_conf.monitor_cgi_url}/main.php', no_auth=True, giveback='raw')
                if versions_raw.status_code == 200:
                    from .CentreonLegacy import CentreonServer as CentreonServerReal
                    self.debug(server='[' + self.get_name() + ']', debug='Loading class LEGACY, Centreon version will be checked later')
                else:
                    self.debug(server='[' + self.get_name() + ']', debug='The URL given is not a Centreon')
                    self.enable = False
                    self.isChecking = False
                    return None

            # kind of mad but helps the servers/__init__.py to detect if there is any other class to be used
            self.ClassServerReal = CentreonServerReal



# encoding: utf-8

import sys
import time
import logging
import requests
from packaging import version

from Nagstamon.helpers import human_readable_duration_from_timestamp
from Nagstamon.config import conf
from Nagstamon.objects import GenericHost, GenericService, Result
from Nagstamon.servers.Generic import GenericServer

class ZabbixLightApi():

    logger = None
    server_name = "Zabbix"
    monitor_url = "http://127.0.0.1/api_jsonrpc.php"
    validate_certs = False
    r_session = None

    zbx_auth = None
    zbx_req_id = 0

    def __init__(self, server_name, monitor_url, validate_certs):

        self.server_name = server_name
        self.monitor_url = monitor_url + "/api_jsonrpc.php"
        self.validate_certs = validate_certs

        #persistent connections
        self.r_session = requests.Session()

        #configure logging
        self.logger = logging.getLogger('ZabbixLightApi_'+server_name)
        console_logger = logging.StreamHandler()
        console_logger_formatter = logging.Formatter('[%(name)s] %(levelname)s - %(message)s')
        console_logger.setFormatter(console_logger_formatter)
        self.logger.addHandler(console_logger)
        if conf.debug_mode is True:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self.logger.debug("ZabbixLightApi START!")
        self.logger.debug("monitor_url = " + self.monitor_url)

    def do_request(self, method, params={}, no_auth=False):
        zabbix_rpc_obj = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
            "id": self.zbx_req_id
        }

        if not no_auth:
            zabbix_rpc_obj["auth"] = self.zbx_auth

        self.zbx_req_id += 1

        self.logger.debug("ZBX: > " + str(zabbix_rpc_obj))

        try:

            response = self.r_session.post(self.monitor_url, json=zabbix_rpc_obj, verify=self.validate_certs)

            #we didnt get HTTP code 200
            if response.status_code != 200:
                raise ZabbixLightApiException("Got return code - " + str(response.status_code))

            #parse response json
            response_json = response.json()
            self.logger.debug("ZBX: < " + str(response_json))

            #there was some kind of error during processing our request
            if "error" in response_json.keys():
                raise ZabbixLightApiException("ZBX: < " + response_json["error"]["data"])

            #zabbix returned garbage
            if "result" not in response_json.keys():
                raise ZabbixLightApiException("ZBX: < no result object in response")

        #all other network related errors
        except Exception as e:
            raise ZabbixLightApiException(e)

        return response_json['result']

    def api_version(self, **options):
        obj = self.do_request('apiinfo.version', options, no_auth=True)
        return obj

    def logged_in(self):
        if self.zbx_auth is None:
            return False
        else:
            is_auth=self.do_request("user.checkAuthentication", {"sessionid": self.zbx_auth}, no_auth=True)
            if is_auth:
                return True
            else:
                self.zbx_auth = None
                return False

    def login(self, username, password):
        self.logger.debug("Login in as " + username)
        # see issue https://github.com/HenriWahl/Nagstamon/issues/1018
        if self.api_version() < '6.4':
            self.zbx_auth = self.do_request('user.login', {'user': username, 'password': password})
        else:
            self.zbx_auth = self.do_request('user.login', {'username': username, 'password': password})

class ZabbixLightApiException(Exception):
    pass

class ZabbixProblemBasedServer(GenericServer):

    TYPE = 'ZabbixProblemBased'
    zlapi = None
    zbx_version = ""

    def __init__(self, **kwds):
        GenericServer.__init__(self, **kwds)

        self.statemap = {
            '0': 'OK',
            '1': 'INFORMATION',
            '2': 'WARNING',
            '3': 'AVERAGE',
            '4': 'HIGH',
            '5': 'DISASTER'}

        # Entries for monitor default actions in context menu
        self.MENU_ACTIONS = []
        # URLs for browser shortlinks/buttons on popup window
        self.BROWSER_URLS = {'monitor': '$MONITOR$',
                             'hosts': '$MONITOR-CGI$/hosts.php?ddreset=1',
                             'services': '$MONITOR-CGI$/zabbix.php?action=problem.view&fullscreen=0&page=1&filter_show=3&filter_set=1',
                             'history':  '$MONITOR-CGI$/zabbix.php?action=problem.view&fullscreen=0&page=1&filter_show=2&filter_set=1'}

        self.username = conf.servers[self.get_name()].username
        self.password = conf.servers[self.get_name()].password
        self.validate_certs = not conf.servers[self.get_name()].ignore_cert

    def _get_status(self):
        """
            Get status from Zabbix Server
        """
        # =========================================
        # problems
        # =========================================
        problems = []
        try:
            #Are we logged in?
            if self.zlapi is None:
                self.zlapi = ZabbixLightApi(server_name=self.name, monitor_url=self.monitor_url, validate_certs=self.validate_certs)

            #zabbix could get an upgrade between checks, we need to check version each time
            self.zbx_version = self.zlapi.do_request("apiinfo.version", {}, no_auth=True)

            #check are we still logged in, if not, relogin
            if not self.zlapi.logged_in():
                self.zlapi.login(self.username, self.password)

            #Get all current problems (trigger based), no need to check acknowledged problems if they are filtered out (load reduce)
            if conf.filter_acknowledged_hosts_services:
                # old versions doesnt support suppressed problems
                if version.parse(self.zbx_version) < version.parse("6.2.0"):
                    problems = self.zlapi.do_request("problem.get", {'recent': False, 'acknowledged': False})
                else:
                    problems = self.zlapi.do_request("problem.get", {'recent': False, 'acknowledged': False, 'suppressed': False})
            else:
                problems = self.zlapi.do_request("problem.get", {'recent': False})

            for problem in problems:

                #get trigger which rose current problem
                trigger = self.zlapi.do_request("trigger.get", {'triggerids': problem['objectid'],
                                                'monitored': True,
                                                'active': True,
                                                'skipDependent': True,
                                                'selectHosts': ['hostid', 'name', 'maintenance_status',
                                                                'available', 'error', 'errors_from',
                                                                'ipmi_available', 'ipmi_error', 'ipmi_errors_from',
                                                                'jmx_available', 'jmx_error', 'jmx_errors_from',
                                                                'snmp_available', 'snmp_error', 'snmp_errors_from'],
                                                'selectItems': ['key_', 'lastclock']})


                #problems on disabled/maintenance/deleted hosts don't have triggers
                #have to do that because of how zabbix housekeeping service work
                #API reports past problems for hosts that no longer exist
                if not trigger:
                    continue

                service_id = problem['eventid']
                host_id = trigger[0]['hosts'][0]['hostid']

                #new host to report, we only need to do that at first problem for that host
                if host_id not in self.new_hosts:
                    self.new_hosts[host_id] = GenericHost()
                    self.new_hosts[host_id].name = trigger[0]['hosts'][0]['name']

                    #host has active maintenance period
                    if trigger[0]['hosts'][0]['maintenance_status'] == "1":
                        self.new_hosts[host_id].scheduled_downtime = True

                    #old api shows host interfaces status in host object
                    if version.parse(self.zbx_version) < version.parse("5.4.0"):

                        #host not available via agent
                        if trigger[0]['hosts'][0].get('available', '0') == "2":
                            self.new_hosts[host_id].status = "DOWN"
                            self.new_hosts[host_id].status_information = trigger[0]['hosts'][0]['error']
                            self.new_hosts[host_id].duration = human_readable_duration_from_timestamp(trigger[0]['hosts'][0]['errors_from'])

                        #host not available via ipmi
                        if trigger[0]['hosts'][0].get('ipmi_available', '0') == "2":
                            self.new_hosts[host_id].status = "DOWN"
                            self.new_hosts[host_id].status_information = trigger[0]['hosts'][0]['ipmi_error']
                            self.new_hosts[host_id].duration = human_readable_duration_from_timestamp(trigger[0]['hosts'][0]['ipmi_errors_from'])

                        #host not available via jmx
                        if trigger[0]['hosts'][0].get('jmx_available', '0') == "2":
                            self.new_hosts[host_id].status = "DOWN"
                            self.new_hosts[host_id].status_information = trigger[0]['hosts'][0]['jmx_error']
                            self.new_hosts[host_id].duration = human_readable_duration_from_timestamp(trigger[0]['hosts'][0]['jmx_errors_from'])

                        #host not available via snmp
                        if trigger[0]['hosts'][0].get('snmp_available', '0') == "2":
                            self.new_hosts[host_id].status = "DOWN"
                            self.new_hosts[host_id].status_information = trigger[0]['hosts'][0]['snmp_error']
                            self.new_hosts[host_id].duration = human_readable_duration_from_timestamp(trigger[0]['hosts'][0]['snmp_errors_from'])

                    #new api shows host interfaces status in hostinterfaces object
                    else:

                        #get all host interfaces
                        hostinterfaces = self.zlapi.do_request("hostinterface.get", {"hostids": host_id})

                        #check them all and mark host as DOWN on first not available interface
                        for hostinterface in hostinterfaces:
                            if hostinterface.get('available', '0') == "2":
                                self.new_hosts[host_id].status = "DOWN"
                                self.new_hosts[host_id].status_information = hostinterface['error']
                                self.new_hosts[host_id].duration = human_readable_duration_from_timestamp(hostinterface['errors_from'])
                                #we stop checking rest of interfaces
                                break

                #service to report
                self.new_hosts[host_id].services[service_id] = GenericService()
                self.new_hosts[host_id].services[service_id].host = trigger[0]['hosts'][0]['name']
                self.new_hosts[host_id].services[service_id].status = self.statemap.get(problem['severity'], problem['severity'])
                self.new_hosts[host_id].services[service_id].duration = human_readable_duration_from_timestamp(problem['clock'])
                self.new_hosts[host_id].services[service_id].name = trigger[0]['items'][0]['key_']
                self.new_hosts[host_id].services[service_id].last_check = time.strftime("%d/%m/%Y %H:%M:%S", time.localtime(int(trigger[0]['items'][0]['lastclock'])))

                #we add opdata to status information just like in zabbix GUI
                if problem["opdata"] != "":
                    self.new_hosts[host_id].services[service_id].status_information = problem['name'] + " (" + problem["opdata"] + ")"
                else:
                    self.new_hosts[host_id].services[service_id].status_information = problem['name']

                #service is acknowledged
                if problem['acknowledged'] == "1":
                    self.new_hosts[host_id].services[service_id].acknowledged = True

        except ZabbixLightApiException:
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        return Result()

    # Disable set_recheck (nosense in Zabbix)
    def set_recheck(self, info_dict):
        pass



# This is part of Sangoma's pysensu library found here: https://github.com/sangoma/pysensu
# and licensed under the MIT License
#
# The MIT License (MIT)

# Copyright (c) 2015 Sangoma Technologies

# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
import logging

import requests
from requests.auth import HTTPBasicAuth

logger = logging.getLogger(__name__)


class SensuAPIException(Exception):
    pass


class SensuAPI:
    def __init__(self, url_base, username=None, password=None, verify=None):
        self._url_base = url_base
        self._session = requests.Session()
        self._session.headers.update({
            'User-Agent': 'PySensu Client v0.9.0'
        })
        self._session.verify = verify
        self.good_status = (200, 201, 202, 204)

        if username and password:
            self._session.auth = HTTPBasicAuth(username, password)
        elif password and not username:
            self._session.headers.update({
                'Authorization': 'token {}'.format(password)
            })

    def _request(self, method, path, **kwargs):
        url = '{}{}'.format(self._url_base, path)
        logger.debug('{} -> {} with {}'.format(method, url, kwargs))

        if method == 'GET':
            resp = self._session.get(url, **kwargs)

        elif method == 'POST':
            resp = self._session.post(url, **kwargs)

        elif method == 'PUT':
            resp = self._session.put(url, **kwargs)

        elif method == 'DELETE':
            resp = self._session.delete(url, **kwargs)
        else:
            raise SensuAPIException(
                'Method {} not implemented'.format(method)
            )

        if resp.status_code in self.good_status:
            logger.debug('{}: {}'.format(
                resp.status_code,
                ''.join(resp.text.split('\n'))[0:80]
            ))
            return resp

        elif resp.status_code == 400:
            logger.error('{}: {}'.format(
                resp.status_code,
                resp.text
            ))
            raise SensuAPIException('API returned "Bad Request"')

        else:
            logger.warning('{}: {}'.format(
                resp.status_code,
                resp.text
            ))
            raise SensuAPIException('API bad response {}: {}'.format(resp.status_code, resp.text))

    """
    Clients ops
    """
    def get_clients(self, limit=None, offset=None):
        """
        Returns a list of clients.
        """
        data = {}
        if limit:
            data['limit'] = limit
        if offset:
            data['offset'] = offset
        result = self._request('GET', '/clients', data=json.dumps(data))
        return result.json()

    def get_client_data(self, client):
        """
        Returns a client.
        """
        data = self._request('GET', '/clients/{}'.format(client))
        return data.json()

    def get_client_history(self, client):
        """
        Returns the history for a client.
        """
        data = self._request('GET', '/clients/{}/history'.format(client))
        return data.json()

    def delete_client(self, client):
        """
        Removes a client, resolving its current events. (delayed action)
        """
        self._request('DELETE', '/clients/{}'.format(client))
        return True

    """
    Events ops
    """
    def get_events(self):
        """
        Returns the list of current events.
        """
        data = self._request('GET', '/events')
        return data.json()

    def get_all_client_events(self, client):
        """
        Returns the list of current events for a given client.
        """
        data = self._request('GET', '/events/{}'.format(client))
        return data.json()

    def get_event(self, client, check):
        """
        Returns an event for a given client & check name.
        """
        data = self._request('GET', '/events/{}/{}'.format(client, check))
        return data.json()

    def delete_event(self, client, check):
        """
        Resolves an event for a given check on a given client. (delayed action)
        """
        self._request('DELETE', '/events/{}/{}'.format(client, check))
        return True

    def post_event(self, client, check):
        """
        Resolves an event. (delayed action)
        """
        self._request('POST', '/resolve', json={'client': client, 'check': check})
        return True

    """
    Checks ops
    """
    def get_checks(self):
        """
        Returns the list of checks.
        """
        data = self._request('GET', '/checks')
        return data.json()

    def get_check(self, check):
        """
        Returns a check.
        """
        data = self._request('GET', '/checks/{}'.format(check))
        return data.json()

    def post_check_request(self, check, subscribers, dc=None):
        """
        Issues a check execution request.
        """
        data = {
            'check': check,
            'subscribers': [subscribers]
        }
        if dc is not None:
            data['dc'] = dc
        self._request('POST', '/request', json=data)
        return True

    """
    Silenced API ops
    """
    def get_silenced(self, limit=None, offset=None):
        """
        Returns a list of silence entries.
        """
        data = {}
        if limit:
            data['limit'] = limit
        if offset:
            data['offset'] = offset
        result = self._request('GET', '/silenced', json=data)
        return result.json()

    def post_silence_request(self, kwargs):
        """
        Create a silence entry.
        """
        self._request('POST', '/silenced', json=kwargs)
        return True

    def clear_silence(self, kwargs):
        """
        Clear a silence entry.
        """
        self._request('POST', '/silenced/clear', json=kwargs)
        return True

    """
    Aggregates ops
    """
    def get_aggregates(self):
        """
        Returns the list of named aggregates.
        """
        data = self._request('GET', '/aggregates')
        return data.json()

    def get_aggregate_check(self, check, age=None):
        """
        Returns the list of aggregates for a given check
        """
        data = {}
        if age:
            data['max_age'] = age

        result = self._request('GET', '/aggregates/{}'.format(check), json=data)
        return result.json()

    def delete_aggregate(self, check):
        """
        Deletes all aggregate data for a named aggregate
        """
        self._request('DELETE', '/aggregates/{}'.format(check))
        return True

    """
    Status ops
    """
    def get_info(self):
        """
        Returns information on the API.
        """
        data = self._request('GET', '/info')
        return data.json()

    def get_health(self, consumers=2, messages=100):
        """
        Returns health information on transport & Redis connections.
        """
        data = {'consumers': consumers, 'messages': messages}

        try:
            self._request('GET', '/health', json=data)
            return True
        except SensuAPIException:
            return False

    """
    Results ops
    """
    def get_all_client_results(self):
        """
        Returns the list of results.
        """
        data = self._request('GET', '/results')
        return data.json()

    def get_results(self, client):
        """
        Returns a result.
        """
        data = self._request('GET', '/results/{}'.format(client))
        return data.json()

    def get_result(self, client, check):
        """
        Returns an event for a given client & result name.
        """
        data = self._request('GET', '/results/{}/{}'.format(client, check))
        return data.json()

    def delete_result(self, client, check):
        """
        Deletes an check result data for a given check on a given client.
        """
        self._request('DELETE', '/results/{}/{}'.format(client, check))
        return True

    def post_result_data(self, client, check, output, status):
        """
        Posts check result data.
        """
        data = {
            'source': client,
            'name': check,
            'output': output,
            'status': status,
        }
        self._request('POST', '/results', json=data)
        return True

    """
    Stashes ops
    """
    def get_stashes(self):
        """
        Returns a list of stashes.
        """
        data = self._request('GET', '/stashes')
        return data.json()

    def create_stash(self, payload, path=None):
        """
        Create a stash. (JSON document)
        """
        if path:
            self._request('POST', '/stashes/{}'.format(path),
                          json=payload)
        else:
            self._request('POST', '/stashes', json=payload)
        return True

    def delete_stash(self, path):
        """
        Delete a stash. (JSON document)
        """
        self._request('DELETE', '/stashes/{}'.format(path))
        return True

    """
    Subscriptions ops (not directly in the Sensu API)
    """
    def get_subscriptions(self, nodes=[]):
        """
        Returns all the channels where (optionally specified) nodes are subscribed
        """
        if len(nodes) > 0:
            data = [node for node in self.get_clients() if node['name'] in nodes]
        else:
            data = self.get_clients()
        channels = []
        for client in data:
            if 'subscriptions' in client:
                if isinstance(client['subscriptions'], list):
                    for channel in client['subscriptions']:
                        if channel not in channels:
                            channels.append(channel)
                else:
                    if client['subscriptions'] not in channels:
                        channels.append(client['subscriptions'])
        return channels

    def get_subscriptions_channel(self, search_channel):
        """
        Return all the nodes that are subscribed to the specified channel
        """
        data = self.get_clients()
        clients = []
        for client in data:
            if 'subscriptions' in client:
                if isinstance(client['subscriptions'], list):
                    if search_channel in client['subscriptions']:
                        clients.append(client['name'])
                else:
                    if search_channel == client['subscriptions']:
                        clients.append(client['name'])
        return clients



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# The initial implementation was contributed to the Nagstamon project
# by tribe29 GmbH.

from datetime import datetime
import copy
import html
import json
import socket
import sys
import urllib.request, urllib.parse, urllib.error
import time

import tzlocal

from Nagstamon.objects import (GenericHost,
                               GenericService,
                               Result)
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.helpers import webbrowser_open
from Nagstamon.config import conf


class MultisiteError(Exception):
    def __init__(self, terminate, result):
        self.terminate = terminate
        self.result    = result


class MultisiteServer(GenericServer):
    """
       special treatment for Checkmk Multisite JSON API
    """
    TYPE = 'Checkmk Multisite'

    # URLs for browser shortlinks/buttons on popup window
    BROWSER_URLS= { 'monitor': '$MONITOR$',
                    'hosts': '$MONITOR$/index.py?start_url=view.py?view_name=hostproblems',
                    'services': '$MONITOR$/index.py?start_url=view.py?view_name=svcproblems',
                    'history': '$MONITOR$/index.py?start_url=view.py?view_name=events'}

    def __init__(self, **kwds):
        GenericServer.__init__(self, **kwds)

        # Prepare all urls needed by nagstamon -
        self.urls = {}
        self.statemap = {}

        # Entries for monitor default actions in context menu
        self.MENU_ACTIONS = ['Monitor', 'Recheck', 'Acknowledge', 'Downtime']

        # flag for newer cookie authentication
        self.cookie_auth = False


    def init_http(self):
        # general initialization
        if not self.session:
            GenericServer.init_http(self)

        # Fix eventually missing tailing '/' in url
        if self.monitor_url.endswith('/'):
            self.monitor_url.rstrip('/')

        # Add /check_mk if not already existent - makes setting URL simpler
        if not self.monitor_url.endswith('/check_mk'):
            self.monitor_url += '/check_mk'

        # Prepare all urls needed by nagstamon if not yet done
        if len(self.urls) == len(self.statemap):
            self.urls = {
              'api_services':    self.monitor_url + '/view.py?view_name={0}&output_format=python&lang=&limit=hard'.\
                                                                          format(self.checkmk_view_services),
              'human_services':  self.monitor_url + '/index.py?%s' % \
                                                   urllib.parse.urlencode({'start_url': 'view.py?view_name={0}'.\
                                                                          format(self.checkmk_view_services)}),
              'human_service':   self.monitor_url + '/index.py?%s' %
                                                   urllib.parse.urlencode({'start_url': 'view.py?view_name=service'}),

              'api_hosts':       self.monitor_url + '/view.py?view_name={0}&output_format=python&lang=&limit=hard'.\
                                                                          format(self.checkmk_view_hosts),
              'human_hosts':     self.monitor_url + '/index.py?%s' %
                                                   urllib.parse.urlencode({'start_url': 'view.py?view_name={0}'.\
                                                                           format(self.checkmk_view_services)}),
              'human_host':      self.monitor_url + '/index.py?%s' %
                                                   urllib.parse.urlencode({'start_url': 'view.py?view_name=hoststatus'}),
              # URLs do not need pythonic output because since werk #0766 API does not work with transid=-1 anymore
              # thus access to normal webinterface is used
              'api_host_act':    self.monitor_url + '/view.py?_transid=-1&_do_actions=yes&_do_confirm=Yes!&view_name=hoststatus&filled_in=actions&lang=',
              'api_service_act': self.monitor_url + '/view.py?_transid=-1&_do_actions=yes&_do_confirm=Yes!&view_name=service&filled_in=actions&lang=',

              # https://checkmk.nkrn.de/cmk/check_mk/view.py?_csrf_token=6769fe9d-44e4-4638-b7a9-2fb7c8d09b75&filled_in=actions&_transid=1768551477%2F300479071&_do_actions=yes&actions=yes&_resched_spread=5&_resched_checks=Reschedule&view_name=svcproblems&lang=
              'api_svcprob_act': self.monitor_url + '/view.py',
              'api_svcprob_act_orig': self.monitor_url + '/view.py?_transid=-1&_do_actions=yes&_do_confirm=Yes!&view_name=svcproblems&filled_in=actions&lang=',

              'human_events':    self.monitor_url + '/index.py?%s' %
                                                   urllib.parse.urlencode({'start_url': 'view.py?view_name=events'}),
              'omd_host_downtime': self.monitor_url + '/api/1.0/domain-types/downtime/collections/host',
              'omd_svc_downtime': self.monitor_url + '/api/1.0/domain-types/downtime/collections/service',
              'recheck':         self.monitor_url + '/ajax_reschedule.py?_ajaxid=0',
              'omd_version':         self.monitor_url + '/api/1.0/version',
              'transid':         self.monitor_url + '/view.py?actions=yes&filled_in=actions&host=$HOST$&service=$SERVICE$&view_name=service'
            }

            self.statemap = {
                'UNREACH': 'UNREACHABLE',
                'CRIT':    'CRITICAL',
                'WARN':    'WARNING',
                'UNKN':    'UNKNOWN',
                'PEND':    'PENDING',
            }

        # Function overrides for Checkmk 2.3+
        version = self._get_version()
        if version >= [2, 3]:
            self._set_downtime = self._set_downtime_since_2_3
            self._set_recheck = self._set_recheck_since_2_3

        if self.authentication != 'web':
            if self.cookie_auth and not self.refresh_authentication:
                # get cookie to access Checkmk web interface
                if 'cookies' in dir(self.session):
                    if len(self.session.cookies) == 0:
                        # if no cookie yet login
                        self._get_cookie_login()
                elif self.session == None:
                    # if no cookie yet login
                    self._get_cookie_login()
            elif self.cookie_auth and \
                    self.refresh_authentication:
                if self.session is None:
                    self.session = self.create_session()

                    # force re-auth
                    self._get_cookie_login()


    def init_config(self):
        """
        dummy init_config, called at thread start, not really needed here, just omit extra properties
        """
        pass


    def _is_auth_in_cookies(self):
        """
        check if there is any valid auth session in cookies which has the name 'auth_<monitor_name>'
        """
        if self.session:
            for cookie in self.session.cookies:
                if cookie.name.startswith('auth_'):
                    return True
        return False

    def _get_url(self, url):
        result = self.fetch_url(url, 'raw')
        content, error, status_code = result.result, result.error, result.status_code

        if error != '' or status_code >= 400:
            raise MultisiteError(True, Result(result=content,
                                              error=error,
                                              status_code=status_code))

        if content.startswith('WARNING:'):
            c = content.split('\n')

            # Print non ERRORS to the log in debug mode
            self.debug(server=self.get_name(), debug=c[0])

            raise MultisiteError(False, Result(result='\n'.join(c[1:]),
                                               error=c[0],
                                               status_code=status_code))

        elif content.startswith('ERROR:'):
            raise MultisiteError(True, Result(result=content,
                                              error=content,
                                              status_code=status_code))

        # in case of auth problem enable GUI auth part in popup
        #if self.cookie_auth and self.session is not None:
        #    if not self._is_auth_in_cookies():
        #        self.refresh_authentication = True
        #        return ''

        # looks like cookieauth
        elif content.startswith('<') or\
                '<!DOCTYPE html>' in content:
            self.cookie_auth = True
            # if first attempt login and then try to get data again
            if not self._is_auth_in_cookies():
                if self.authentication != 'web':
                    self._get_cookie_login()
                result = self.fetch_url(url, 'raw')
                content, error = result.result, result.error
                if content.startswith('<') or\
                '<!DOCTYPE html>' in content:
                    return ''

        # if finally still some <HTML> is sent this looks like a new login due to password change
        if content.startswith('<') or\
                '<!DOCTYPE html>' in content:
            self.refresh_authentication = True
            return ''

        return eval(content)

    def _get_cookie_login(self):
        """
        login on cookie monitor site
        """
        # put all necessary data into url string
        login_data = {'_username': self.get_username(),
                     '_password': self.get_password(),
                     '_login': '1',
                     '_origtarget' : '',
                     'filled_in' :' login'}
        # get cookie from login page via url retrieving as with other urls
        try:
            # login and get cookie
            self.fetch_url(self.monitor_url + '/login.py', cgi_data=login_data, multipart=True)
        except:
            self.error(sys.exc_info())


    def _get_status(self):
        """
        Get status from Checkmk Server
        """

        ret = Result()

        # Create URLs for the configured filters
        url_params = ''

        if self.force_authuser:
            url_params += "&force_authuser=1"

        url_params += '&is_host_acknowledged=-1&is_service_acknowledged=-1'
        url_params += '&is_host_notifications_enabled=-1&is_service_notifications_enabled=-1'
        url_params += '&is_host_active_checks_enabled=-1&is_service_active_checks_enabled=-1'
        url_params += '&host_scheduled_downtime_depth=-1&is_in_downtime=-1'

        try:
            response = []
            try:
                response = self._get_url(self.urls['api_hosts'] + url_params)
            except MultisiteError as e:
                if e.terminate:
                    return e.result

            if response == '':
                return Result(result='',
                              error='Login failed',
                              status_code=401)

            for row in response[1:]:
                host= dict(list(zip(copy.deepcopy(response[0]), copy.deepcopy(row))))
                n = {
                    'host':               host['host'],
                    'status':             self.statemap.get(host['host_state'], host['host_state']),
                    'last_check':         host['host_check_age'],
                    'duration':           host['host_state_age'],
                    'status_information': html.unescape(host['host_plugin_output'].replace('\n', ' ')),
                    'attempt':            host['host_attempt'],
                    'site':               host['sitename_plain'],
                    'address':            host['host_address']
                }

                # host objects contain service objects
                if n['host'] not in self.new_hosts:
                    new_host = n['host']
                    self.new_hosts[new_host] = GenericHost()
                    self.new_hosts[new_host].name = n['host']
                    self.new_hosts[new_host].server = self.name
                    self.new_hosts[new_host].status = n['status']
                    self.new_hosts[new_host].last_check = n['last_check']
                    self.new_hosts[new_host].duration = n['duration']
                    self.new_hosts[new_host].attempt = n['attempt']
                    self.new_hosts[new_host].status_information= html.unescape(n['status_information'].replace('\n', ' '))
                    self.new_hosts[new_host].site = n['site']
                    self.new_hosts[new_host].address = n['address']

                    # transisition to Checkmk 1.1.10p2
                    if 'host_in_downtime' in host:
                        if host['host_in_downtime'] == 'yes':
                            self.new_hosts[new_host].scheduled_downtime = True
                    if 'host_acknowledged' in host:
                        if host['host_acknowledged'] == 'yes':
                            self.new_hosts[new_host].acknowledged = True
                    if 'host_notifications_enabled' in host:
                        if host['host_notifications_enabled'] == 'no':
                            self.new_hosts[new_host].notifications_disabled = True

                    # hard/soft state for later filter evaluation
                    real_attempt, max_attempt = self.new_hosts[new_host].attempt.split('/')
                    if real_attempt != max_attempt:
                        self.new_hosts[new_host].status_type = 'soft'
                    else:
                        self.new_hosts[new_host].status_type = 'hard'

            del response

        except:
            import traceback
            traceback.print_exc(file=sys.stdout)

            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        # Add filters to the url which should only be applied to the service request
        if conf.filter_services_on_unreachable_hosts:
            # thanks to https://github.com/HenriWahl/Nagstamon/issues/510
            url_params += '&hst0=On&hst1=On'

        # services
        try:
            response = []
            try:
                response = self._get_url(self.urls['api_services'] + url_params)
            except MultisiteError as e:
                if e.terminate:
                    return e.result
                else:
                    response = copy.deepcopy(e.result.content)
                    ret = copy.deepcopy(e.result)

            for row in response[1:]:
                service = dict(list(zip(copy.deepcopy(response[0]), copy.deepcopy(row))))
                n = {
                    'host':               service['host'],
                    'service':            service['service_description'],
                    'status':             self.statemap.get(service['service_state'], service['service_state']),
                    'last_check':         service['svc_check_age'],
                    'duration':           service['svc_state_age'],
                    'attempt':            service['svc_attempt'],
                    'status_information': html.unescape(service['svc_plugin_output'].replace('\n', ' ')),
                    # Checkmk passive services can be re-scheduled by using the Checkmk service
                    'passiveonly':        service['svc_is_active'] == 'no' and not service['svc_check_command'].startswith('check_mk'),
                    'flapping':           service['svc_flapping'] == 'yes',
                    'site':               service['sitename_plain'],
                    'address':            service['host_address'],
                    'command':            service['svc_check_command'],
                }

                # host objects contain service objects
                if n['host'] not in self.new_hosts:
                    self.new_hosts[n['host']] = GenericHost()
                    self.new_hosts[n['host']].name = n['host']
                    self.new_hosts[n['host']].status = 'UP'
                    self.new_hosts[n['host']].site = n['site']
                    self.new_hosts[n['host']].address = n['address']
                # if a service does not exist create its object
                if n['service'] not in self.new_hosts[n['host']].services:
                    new_service = n['service']
                    self.new_hosts[n['host']].services[new_service] = GenericService()
                    self.new_hosts[n['host']].services[new_service].host = n['host']
                    self.new_hosts[n['host']].services[new_service].server = self.name
                    self.new_hosts[n['host']].services[new_service].name = n['service']
                    self.new_hosts[n['host']].services[new_service].status = n['status']
                    self.new_hosts[n['host']].services[new_service].last_check = n['last_check']
                    self.new_hosts[n['host']].services[new_service].duration = n['duration']
                    self.new_hosts[n['host']].services[new_service].attempt = n['attempt']
                    self.new_hosts[n['host']].services[new_service].status_information = n['status_information'].strip()
                    self.new_hosts[n['host']].services[new_service].passiveonly = n['passiveonly']
                    self.new_hosts[n['host']].services[new_service].flapping = n['flapping']
                    self.new_hosts[n['host']].services[new_service].site = n['site']
                    self.new_hosts[n['host']].services[new_service].address = n['address']
                    self.new_hosts[n['host']].services[new_service].command = n['command']

                    # transition to Checkmk 1.1.10p2
                    if 'svc_in_downtime' in service:
                        if service['svc_in_downtime'] == 'yes':
                            self.new_hosts[n['host']].services[new_service].scheduled_downtime = True
                    if 'svc_acknowledged' in service:
                        if service['svc_acknowledged'] == 'yes':
                            self.new_hosts[n['host']].services[new_service].acknowledged = True
                    if 'svc_flapping' in service:
                        if service['svc_flapping'] == 'yes':
                            self.new_hosts[n['host']].services[new_service].flapping = True
                    if 'svc_notifications_enabled' in service:
                        if service['svc_notifications_enabled'] == 'no':
                            self.new_hosts[n['host']].services[new_service].notifications_disabled = True
                    if 'host_in_downtime' in service:
                        if service['host_in_downtime'] == 'yes':
                            self.new_hosts[n['host']].scheduled_downtime = True

                    # hard/soft state for later filter evaluation
                    real_attempt, max_attempt = self.new_hosts[n['host']].services[new_service].attempt.split('/')
                    if real_attempt != max_attempt:
                        self.new_hosts[n['host']].services[new_service].status_type = 'soft'
                    else:
                        self.new_hosts[n['host']].services[new_service].status_type = 'hard'

            del response

        except:
            import traceback
            traceback.print_exc(file=sys.stdout)

            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            return Result(result=copy.deepcopy(result), error=copy.deepcopy(error))

        del url_params

        return ret


    def open_monitor(self, host, service=''):
        """
        open monitor from treeview context menu
        """

        if service == '':
            url = self.urls['human_host'] + urllib.parse.urlencode({'x': 'site='+self.hosts[host].site+'&host='+host}).replace('x=', '%26')
        else:
            url = self.urls['human_service'] + urllib.parse.urlencode({'x': 'site='+self.hosts[host].site+'&host='+host+'&service='+service}).replace('x=', '%26')

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, service=service, debug='Open host/service monitor web page ' + url)
        webbrowser_open(url)


    def get_host(self, host):
        """
        find out ip or hostname of given host to access hosts/devices which do not appear in DNS but
        have their ip saved in Nagios
        """

        # the fastest method is taking hostname as used in monitor
        if conf.connect_by_host == True or host == '':
            return Result(result=host)

        ip = ''

        try:
            if host in self.hosts:
                ip = self.hosts[host].address

            if conf.debug_mode:
                self.debug(server=self.get_name(), host=host, debug ='IP of %s:' % (host) + ' ' + ip)

            if conf.connect_by_dns:
                try:
                    address = socket.gethostbyaddr(ip)[0]
                except:
                    address = ip
            else:
                address = ip
        except:
            result, error = self.error(sys.exc_info())
            return Result(result=result, error=error)

        return Result(result=address)


    def get_start_end(self, host):
        return time.strftime('%Y-%m-%d %H:%M'), time.strftime('%Y-%m-%d %H:%M', time.localtime(time.time() + 7200))


    def _action(self, site, host, service, specific_params):
        params = {
            'site':    self.hosts[host].site,
            'host':    host,
        }
        params.update(specific_params)

        # decide about service or host url
        if service != '':
            url = self.urls['api_service_act']
        else:
            url = self.urls['api_host_act']

        # set service
        params['service'] = service

        # get current transid
        transid = self._get_transid(host, service)
        url = url.replace('?_transid=-1&', f'?_transid={transid}&')

        if conf.debug_mode:
            self.debug(server=self.get_name(), host=host, debug =f'Submitting action: {url} & {urllib.parse.urlencode(params)}')

        # apply action
        self.fetch_url(url + '&' + urllib.parse.urlencode(params))

    def _set_downtime(self, host, service, author, comment, fixed, start_time, end_time, hours, minutes):
        try:
            # might be more sophisticated, especially if there is a localized Checkmk web interface
            from_date, from_time = start_time.split(' ')
            from_year, from_month, from_day = from_date.split('-')
            from_hour, from_min = from_time.split(':')
            to_date, to_time = end_time.split(' ')
            to_year, to_month, to_day = to_date.split('-')
            to_hour, to_min = to_time.split(':')

            # let's try to push downtime info in all variants to server - somewhat holzhammery but well...
            params = {
                '_down_comment': author == self.username and comment or '%s: %s' % (author, comment),
                '_down_flexible': fixed == 0 and 'on' or '',
                '_down_custom': 'Custom+time+range',
                '_down_from_date': from_date,
                '_down_from_time': from_time,
                '_down_to_date': to_date,
                '_down_to_time': to_time,
                '_down_duration': '%s:%s' % (hours, minutes),
                '_down_from_year': from_year,
                '_down_from_month': from_month,
                '_down_from_day': from_day,
                '_down_from_hour': from_hour,
                '_down_from_min': from_min,
                '_down_from_sec': '00',
                '_down_to_year': to_year,
                '_down_to_month': to_month,
                '_down_to_day': to_day,
                '_down_to_hour': to_hour,
                '_down_to_min': to_min,
                '_down_to_sec': '00',
                'actions': 'yes'
            }
            # service needs extra parameter
            if service:
                params['_do_confirm_service_downtime'] = 'Schedule+downtime+for+1+service'
            self._action(self.hosts[host].site, host, service, params)
        except:
            if conf.debug_mode:
                self.debug(server=self.get_name(), host=host,
                           debug='Invalid start/end date/time given')


    def _set_downtime_since_2_3(self, host, service, author, comment, fixed, start_time, end_time, hours, minutes):
        """
        _set_downtime function for Checkmk version 2.3+
        """
        try:
            # Headers required for Checkmk API
            headers = {
                "Authorization": f"Bearer {self.username} {self.password}",
                "Content-Type": "application/json",
                "Accept": "application/json",
            }

            # Only timezone aware dates are allowed
            iso_start_time = datetime.strptime(start_time, "%Y-%m-%d %H:%M").replace(tzinfo=tzlocal.get_localzone()).isoformat()
            iso_end_time = datetime.strptime(end_time, "%Y-%m-%d %H:%M").replace(tzinfo=tzlocal.get_localzone()).isoformat()
            # Set parameters for host downtimes
            url = self.urls["omd_host_downtime"]
            params = {
                "start_time": iso_start_time,
                "end_time": iso_end_time,
                "comment": author == self.username and comment or "%s: %s" % (author, comment),
                "downtime_type": "host",
                "host_name": host,
            }

            # Downtime type is "flexible" if "duration" is set
            if fixed == 0:
                params['duration'] = hours * 60 + minutes
            # Parameter overrides for service downtimes
            if service:
                url = self.urls['omd_svc_downtime']
                params['downtime_type'] = 'service'
                params['service_descriptions'] = [service]
            result = self.fetch_url(url, headers=headers, cgi_data=json.dumps(params))
        except Exception as error:
            if conf.debug_mode:
                self.debug(server=self.get_name(), host=host,
                           debug='Invalid start/end date/time given')


    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent, all_services=None):
        specific_params = {
            '_acknowledge':    'Acknowledge',
            '_ack_sticky':     sticky == 1 and 'on' or '',
            '_ack_notify':     notify == 1 and 'on' or '',
            '_ack_persistent': persistent == 1 and 'on' or '',
            '_ack_comment':    author == self.username and comment or '%s: %s' % (author, comment)
        }
        self._action(self.hosts[host].site, host, service, specific_params)

        # acknowledge all services on a host when told to do so
        if all_services:
            for s in all_services:
                self._action(self.hosts[host].site, host, s, specific_params)


    def _set_recheck(self, host, service):
        specific_params = {
            '_resched_checks': 'Reschedule active checks',
            '_resched_spread':  '0'
        }
        self._action(self.hosts[host].site, host, service, specific_params)


    def _set_recheck_since_2_3(self, host, service):
        """
        _set_recheck function for Checkmk version 2.3+
        """
        csrf_token = self._get_csrf_token(host, service)
        data = {
            "site": self.hosts[host].site,
            "host": host,
            "service": service,
            "wait_svc": service,
            "_csrf_token": csrf_token,
        }
        result = self.fetch_url(self.urls["recheck"], cgi_data=data)

    def _get_transid(self, host, service):
        """
        get transid for an action
        """
        # since Checkmk 2.0 it seems to be a problem if service is empty so fill it with a definitively existing one
        if not service:
            service = 'PING'
        transid = self.fetch_url(self.urls['transid'].replace('$HOST$', host).replace('$SERVICE$', service.replace(' ', '+')),
                                'obj').result.find(attrs={'name' : '_transid'})['value']
        return transid


    def _get_csrf_token(self, host, service):
        """
        get csrf token for the session
        """
        # since Checkmk 2.0 it seems to be a problem if service is empty so fill it with a definitively existing one
        if not service:
            service = "PING"
        csrf_token = self.fetch_url(self.urls["transid"].replace("$HOST$", host).replace("$SERVICE$", service.replace(" ", "+")),
                                    "obj").result.find(attrs={"name": ["csrf_token", "_csrf_token"]})["value"]
        return csrf_token


    def _get_version(self):
        """
        get version of OMD Checkmk as [major_version, minor_version]
        """
        try:
            # need authentication to access /version api
            version = [int(x) for x in self.fetch_url(self.urls['omd_version'], giveback='json').result['versions']['checkmk'].split('.')[:2]]
        # If /version api is not supported, return the lowest non-negative pair
        except:
            version = [0, 0]
        return version



import sys
import json
import re

from datetime import datetime, timedelta, timezone

from Nagstamon.config import conf
from Nagstamon.objects import (GenericHost, Result)
from Nagstamon.servers.Generic import GenericServer
from Nagstamon.helpers import webbrowser_open

from .helpers import (start_logging,
                      get_duration,
                      convert_timestring_to_utc,
                      detect_from_labels)

from .alertmanagerservice import AlertmanagerService

# TODO: support debug level switching while running
log = start_logging('alertmanager', conf.debug_mode)

class AlertmanagerServer(GenericServer):
    """
    special treatment for alertmanager API
    """
    TYPE = 'Alertmanager'

    # alertmanager actions are limited to visiting the monitor for now
    MENU_ACTIONS = ['Monitor', 'Downtime', 'Acknowledge']
    BROWSER_URLS = {
        'monitor':  '$MONITOR$/#/alerts',
        'hosts':    '$MONITOR$/#/alerts',
        'services': '$MONITOR$/#/alerts',
        'history':  '$MONITOR$/#/alerts'
    }

    API_PATH_ALERTS = "/api/v2/alerts?inhibited=false"
    API_PATH_SILENCES = "/api/v2/silences"
    API_FILTERS = '&filter='

    # vars specific to alertmanager class
    map_to_hostname = ''
    map_to_servicename = ''
    map_to_status_information = ''
    map_to_critical = ''
    map_to_warning = ''
    map_to_down = ''
    map_to_unknown = ''
    map_to_ok = ''
    name = ''
    alertmanager_filter = ''


    def init_http(self):
        """
        things to do if HTTP is not initialized
        """
        GenericServer.init_http(self)

        # prepare for JSON
        if self.session is not None:
            self.session.headers.update({'Accept': 'application/json',
                                         'Content-Type': 'application/json'})


    def init_config(self):
        """
        dummy init_config, called at thread start
        """


    def get_start_end(self, host):
        """
        Set a default of starttime of "now" and endtime is "now + 24 hours"
        directly from web interface
        """
        start = datetime.now()
        end = datetime.now() + timedelta(hours=24)

        return (str(start.strftime("%Y-%m-%d %H:%M:%S")),
                str(end.strftime("%Y-%m-%d %H:%M:%S")))

    def map_severity(self, the_severity):
        """Maps a severity

        Args:
            the_severity (str): The severity that should be mapped

        Returns:
            str: The matched Nagstamon severity
        """
        if the_severity in self.map_to_unknown.split(','):
            return "UNKNOWN"
        if the_severity in self.map_to_critical.split(','):
            return "CRITICAL"
        if the_severity in self.map_to_warning.split(','):
            return "WARNING"
        if the_severity in self.map_to_down.split(','):
            return "DOWN"
        if the_severity in self.map_to_ok.split(','):
            return "OK"
        return the_severity.upper()

    def _process_alert(self, alert):
        result = {}

        # alertmanager specific extensions
        generator_url = alert.get("generatorURL", {})
        fingerprint = alert.get("fingerprint", {})
        log.debug("processing alert with fingerprint '%s':", fingerprint)

        labels = alert.get("labels", {})
        state = alert.get("status", {"state": "active"})["state"]
        severity = self.map_severity(labels.get("severity", "unknown"))

        # skip alerts with none severity
        if severity == "NONE":
            log.debug("[%s]: detected detected state '%s' and severity '%s' from labels \
                      -> skipping alert", fingerprint, state, severity)
            return False
        log.debug("[%s]: detected detected state '%s' and severity '%s' from labels",
                  fingerprint, state, severity)

        hostname = detect_from_labels(labels,self.map_to_hostname,"unknown")
        hostname = re.sub(':[0-9]+', '', hostname)
        log.debug("[%s]: detected hostname from labels: '%s'", fingerprint, hostname)

        servicename = detect_from_labels(labels,self.map_to_servicename,"unknown")
        log.debug("[%s]: detected servicename from labels: '%s'", fingerprint, servicename)

        if "status" in alert:
            attempt = alert["status"].get("state", "unknown")
        else:
            attempt = "unknown"

        if attempt == "suppressed":
            scheduled_downtime = True
            acknowledged = True
            log.debug("[%s]: detected status: '%s' -> interpreting as silenced",
                      fingerprint, attempt)
        else:
            scheduled_downtime = False
            acknowledged = False
            log.debug("[%s]: detected status: '%s'", fingerprint, attempt)

        duration = str(get_duration(alert["startsAt"]))

        annotations = alert.get("annotations", {})
        status_information = detect_from_labels(annotations,self.map_to_status_information,'')

        result['host'] = str(hostname)
        result['name'] = servicename
        result['server'] = self.name
        result['status'] = severity
        result['labels'] = labels
        result['last_check'] = str(get_duration(alert["updatedAt"]))
        result['attempt'] = attempt
        result['scheduled_downtime'] = scheduled_downtime
        result['acknowledged'] = acknowledged
        result['duration'] = duration
        result['generatorURL'] = generator_url
        result['fingerprint'] = fingerprint
        result['status_information'] = status_information

        return result


    def _get_status(self):
        """
        Get status from alertmanager Server
        """

        log.debug("detection config (map_to_status_information): '%s'",
                  self.map_to_status_information)
        log.debug("detection config (map_to_hostname): '%s'",
                  self.map_to_hostname)
        log.debug("detection config (map_to_servicename): '%s'",
                  self.map_to_servicename)
        log.debug("detection config (alertmanager_filter): '%s'",
                  self.alertmanager_filter)
        log.debug("severity config (map_to_unknown): '%s'",
                  self.map_to_unknown)
        log.debug("severity config (map_to_critical): '%s'",
                  self.map_to_critical)
        log.debug("severity config (map_to_warning): '%s'",
                  self.map_to_warning)
        log.debug("severity config (map_to_down): '%s'",
                  self.map_to_down)
        log.debug("severity config (map_to_ok): '%s'",
                  self.map_to_ok)

        # get all alerts from the API server
        try:
            if self.alertmanager_filter != '':
                result = self.fetch_url(self.monitor_url + self.API_PATH_ALERTS + self.API_FILTERS
                                        + self.alertmanager_filter, giveback="raw")
            else:
                result = self.fetch_url(self.monitor_url + self.API_PATH_ALERTS,
                                        giveback="raw")

            if result.status_code == 200:
                log.debug("received status code '%s' with this content in result.result: \n\
                           ---------------------------------------------------------------\n\
                           %s\
                           ---------------------------------------------------------------",
                           result.status_code, result.result)
            else:
                log.error("received status code '%s'", result.status_code)

            try:
                data = json.loads(result.result)
            except json.decoder.JSONDecodeError:
                data = {}
            error = result.error
            status_code = result.status_code

            # check if any error occured
            errors_occured = self.check_for_error(data, error, status_code)
            if errors_occured is not None:
                return errors_occured

            for alert in data:
                alert_data = self._process_alert(alert)
                if not alert_data:
                    continue

                service = AlertmanagerService()
                service.host = alert_data['host']
                service.name = alert_data['fingerprint']
                service.display_name = alert_data['name']
                service.server = alert_data['server']
                service.status = alert_data['status']
                service.labels = alert_data['labels']
                service.scheduled_downtime = alert_data['scheduled_downtime']
                service.acknowledged = alert_data['acknowledged']
                service.last_check = alert_data['last_check']
                service.attempt = alert_data['attempt']
                service.duration = alert_data['duration']

                service.generator_url = alert_data['generatorURL']
                service.fingerprint = alert_data['fingerprint']

                service.status_information = alert_data['status_information']

                if service.host not in self.new_hosts:
                    self.new_hosts[service.host] = GenericHost()
                    self.new_hosts[service.host].name = str(service.host)
                    self.new_hosts[service.host].server = self.name
                self.new_hosts[service.host].services[service.name] = service

        except Exception as the_exception:
            # set checking flag back to False
            self.isChecking = False
            result, error = self.error(sys.exc_info())
            log.exception(the_exception)
            return Result(result=result, error=error)

        # dummy return in case all is OK
        return Result()

    def open_monitor_webpage(self, host, service):
        """
        open monitor from tablewidget context menu
        """
        webbrowser_open('%s' % (self.monitor_url))

    def open_monitor(self, host, service):
        """
        open monitor for alert
        """
        url = self.monitor_url
        webbrowser_open(url)


    def _set_downtime(self, host, service, author, comment, fixed, start_time,
                      end_time, hours, minutes):

        # Services are keyed by fingerprint internally, but the UI passes display_name.
        # Look up the alert by display_name.
        alert = next((s for s in self.hosts[host].services.values()
                      if s.display_name == service), None)
        if alert is None:
            log.error(f'_set_downtime: service "{service}" not found on host "{host}"')
            return

        # Convert local dates to UTC
        start_time_dt = convert_timestring_to_utc(start_time)
        end_time_dt = convert_timestring_to_utc(end_time)

        # API Spec: https://github.com/prometheus/alertmanager/blob/master/api/v2/openapi.yaml
        silence_data = {}
        silence_data["matchers"] = []
        for name, value in alert.labels.items():
            silence_data["matchers"].append({
                "name": name,
                "value": value,
                "isRegex": False,
                "isEqual": True
            })
        silence_data["startsAt"] = start_time_dt
        silence_data["endsAt"] = end_time_dt
        silence_data["comment"] = comment or "Nagstamon downtime"
        silence_data["createdBy"] = author or "Nagstamon"

        self.fetch_url(self.monitor_url + self.API_PATH_SILENCES, giveback="raw",
                       cgi_data=json.dumps(silence_data))


    # Overwrite function from generic server to add expire_time value
    def set_acknowledge(self, info_dict):
        '''
            different monitors might have different implementations of _set_acknowledge
        '''
        if info_dict['acknowledge_all_services'] is True:
            all_services = info_dict['all_services']
        else:
            all_services = []

        # Make sure expire_time is set
        #if not info_dict['expire_time']:
        #    info_dict['expire_time'] = None

        self._set_acknowledge(info_dict['host'],
                              info_dict['service'],
                              info_dict['author'],
                              info_dict['comment'],
                              info_dict['sticky'],
                              info_dict['notify'],
                              info_dict['persistent'],
                              all_services,
                              info_dict['expire_time'])


    def _post_silence(self, alert, author, comment, starts_at, ends_at):
        """Build and POST a silence payload for the given alert object."""
        silence_data = {
            "matchers": [
                {"name": name, "value": value, "isRegex": False}
                for name, value in alert.labels.items()
            ],
            "startsAt": starts_at,
            "endsAt": ends_at,
            "comment": comment or "Nagstamon silence",
            "createdBy": author or "Nagstamon",
        }
        return self.fetch_url(self.monitor_url + self.API_PATH_SILENCES, giveback="raw",
                              cgi_data=json.dumps(silence_data))

    def _set_acknowledge(self, host, service, author, comment, sticky, notify, persistent,
                         all_services=None, expire_time=None):
        # Services are keyed by fingerprint internally, but the UI passes display_name.
        # Look up the alert by display_name.
        alert = next((s for s in self.hosts[host].services.values()
                      if s.display_name == service), None)
        if alert is None:
            log.error(f'_set_acknowledge: service "{service}" not found on host "{host}"')
            return

        starts_at = datetime.now(timezone.utc).isoformat()
        ends_at = convert_timestring_to_utc(expire_time) if expire_time else starts_at

        self._post_silence(alert, author, comment, starts_at, ends_at)

        if all_services:
            for svc_name in all_services:
                svc_alert = next((s for s in self.hosts[host].services.values()
                                  if s.display_name == svc_name), None)
                if svc_alert is None:
                    log.error(f'_set_acknowledge: service "{svc_name}" not found on host "{host}"')
                    continue
                self._post_silence(svc_alert, author, comment, starts_at, ends_at)



# Nagstamon - Nagios status monitor for your desktop
# Copyright (C) 2008-2026 Henri Wahl <henri@nagstamon.de> et al.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


from Nagstamon.config import conf
from Nagstamon.qui.dialogs.dialog import Dialog
from Nagstamon.qui.qt import (QDateTime,
                              Signal,
                              Slot)
from Nagstamon.servers import SERVER_TYPES

class DialogAcknowledge(Dialog):
    """
    dialog for acknowledging host/service problems
    """

    # store host and service to be used for OK button evaluation
    server = None
    host_list = service_list = []

    # tell worker to acknowledge some troublesome item
    acknowledge = Signal(dict)

    def __init__(self,):
        Dialog.__init__(self, 'dialog_acknowledge')

        self.TOGGLE_DEPS = {
            self.window.input_checkbox_use_expire_time: [self.window.input_datetime_expire_time]
        }

        # still clumsy but better than negating the other server types
        PROMETHEUS_OR_ALERTMANAGER = ['Alertmanager',
                                      'Prometheus']
        NOT_PROMETHEUS_OR_ALERTMANAGER = [x.TYPE for x in SERVER_TYPES.values() if
                                          x.TYPE not in PROMETHEUS_OR_ALERTMANAGER]

        self.VOLATILE_WIDGETS = {
            self.window.input_checkbox_use_expire_time: ['IcingaWeb2', 'Icinga2API', 'IcingaDBWeb'],
            self.window.input_datetime_expire_time: ['IcingaWeb2', 'Icinga2API', 'IcingaDBWeb', 'Alertmanager'],
            self.window.input_checkbox_sticky_acknowledgement: NOT_PROMETHEUS_OR_ALERTMANAGER,
            self.window.input_checkbox_send_notification: NOT_PROMETHEUS_OR_ALERTMANAGER,
            self.window.input_checkbox_persistent_comment: NOT_PROMETHEUS_OR_ALERTMANAGER,
            self.window.input_checkbox_acknowledge_all_services: NOT_PROMETHEUS_OR_ALERTMANAGER
        }

        self.FORCE_DATETIME_EXPIRE_TIME = ['Alertmanager']

    @Slot(object, list, list)
    def initialize(self, server=None, host=[], service=[]):
        # store server, host and service to be used for OK button evaluation
        self.server = server
        self.host_list = host
        self.service_list = service

        self.window.setWindowTitle('Acknowledge hosts and services')
        text = ''

        for i in range(len(self.host_list)):
            if self.service_list[i] == "":
                text += f'Host <b>{self.host_list[i]}</b><br>'
            else:
                text += f'Service <b>{self.service_list[i]}</b> on host <b>{self.host_list[i]}</b><br>'

        self.window.input_label_description.setText(text)

        # default flags of monitor acknowledgement
        self.window.input_checkbox_sticky_acknowledgement.setChecked(conf.defaults_acknowledge_sticky)
        self.window.input_checkbox_send_notification.setChecked(conf.defaults_acknowledge_send_notification)
        self.window.input_checkbox_persistent_comment.setChecked(conf.defaults_acknowledge_persistent_comment)
        self.window.input_checkbox_use_expire_time.setChecked(conf.defaults_acknowledge_expire)
        if len(self.host_list) == 1:
            self.window.input_checkbox_acknowledge_all_services.setChecked(conf.defaults_acknowledge_all_services)
            self.window.input_checkbox_acknowledge_all_services.show()
        else:
            self.window.input_checkbox_acknowledge_all_services.setChecked(False)
            self.window.input_checkbox_acknowledge_all_services.hide()

        # default author + comment
        self.window.input_lineedit_comment.setText(conf.defaults_acknowledge_comment)
        self.window.input_lineedit_comment.setFocus()

        # set default and minimum value for expiry time
        qdatetime = QDateTime.currentDateTime()
        self.window.input_datetime_expire_time.setMinimumDateTime(qdatetime)
        # set default expiry time from configuration
        self.window.input_datetime_expire_time.setDateTime(qdatetime.addSecs(
            conf.defaults_acknowledge_expire_duration_hours * 60 * 60 + conf.defaults_acknowledge_expire_duration_minutes * 60
        ))

        # Show or hide widgets based on server
        if self.server is not None:
            for widget, server_types in self.VOLATILE_WIDGETS.items():
                if self.server.TYPE in server_types:
                    widget.show()
                    self.toggle_toggles()
                else:
                    widget.hide()
            if self.server.TYPE in self.FORCE_DATETIME_EXPIRE_TIME:
                self.window.input_datetime_expire_time.show()

        # Adjust to current size if items are hidden in the menu
        # Otherwise it will get confused and chop off text
        self.window.options_groupbox.adjustSize()
        self.window.adjustSize()

    def ok(self):
        """
        acknowledge miserable host/service
        """
        # create a list of all service of selected host to acknowledge them all
        all_services = list()
        acknowledge_all_services = self.window.input_checkbox_acknowledge_all_services.isChecked()

        if acknowledge_all_services is True:
            for i in self.server.nagitems_filtered["services"].values():
                for s in i:
                    if s.host in self.host_list:
                        all_services.append(s.name)

        if self.window.input_checkbox_use_expire_time.isChecked() or self.server.TYPE in self.FORCE_DATETIME_EXPIRE_TIME:
            # Format used in UI
            # 2019-11-01T18:17:39
            expire_datetime = self.window.input_datetime_expire_time.dateTime().toString("yyyy-MM-ddTHH:mm:ss")
        else:
            expire_datetime = None

        for line_number in range(len(self.host_list)):
            service = self.service_list[line_number]
            host = self.host_list[line_number]

            # send signal to tablewidget worker to care about acknowledging with supplied information
            self.acknowledge.emit({'server': self.server,
                                   'host': host,
                                   'service': service,
                                   'author': self.server.username,
                                   'comment': self.window.input_lineedit_comment.text(),
                                   'sticky': self.window.input_checkbox_sticky_acknowledgement.isChecked(),
                                   'notify': self.window.input_checkbox_send_notification.isChecked(),
                                   'persistent': self.window.input_checkbox_persistent_comment.isChecked(),
                                   'acknowledge_all_services': acknowledge_all_services,
                                   'all_services': all_services,
                                   'expire_time': expire_datetime})
        # call close and macOS dock icon treatment from ancestor
        super().ok()



XK_kra = 0x3a2
XK_kappa = 0x3a2
XK_Rcedilla = 0x3a3
XK_Itilde = 0x3a5
XK_Lcedilla = 0x3a6
XK_Emacron = 0x3aa
XK_Gcedilla = 0x3ab
XK_Tslash = 0x3ac
XK_rcedilla = 0x3b3
XK_itilde = 0x3b5
XK_lcedilla = 0x3b6
XK_emacron = 0x3ba
XK_gcedilla = 0x3bb
XK_tslash = 0x3bc
XK_ENG = 0x3bd
XK_eng = 0x3bf
XK_Amacron = 0x3c0
XK_Iogonek = 0x3c7
XK_Eabovedot = 0x3cc
XK_Imacron = 0x3cf
XK_Ncedilla = 0x3d1
XK_Omacron = 0x3d2
XK_Kcedilla = 0x3d3
XK_Uogonek = 0x3d9
XK_Utilde = 0x3dd
XK_Umacron = 0x3de
XK_amacron = 0x3e0
XK_iogonek = 0x3e7
XK_eabovedot = 0x3ec
XK_imacron = 0x3ef
XK_ncedilla = 0x3f1
XK_omacron = 0x3f2
XK_kcedilla = 0x3f3
XK_uogonek = 0x3f9
XK_utilde = 0x3fd
XK_umacron = 0x3fe
